import uuid


class Entity:
    """ Base class of all objects lives in :class:`World`, including cars, and
    all sensors attached to the car.

    """
    def __init__(self):
        # Identifier for this Entity. Unique given during creation.
        self._id = uuid.uuid4().hex[:6]

        # Actors may be attached to a parent actor that they will follow around.
        self._parent = None

    @property
    def id(self):
        """ The identifier of this entity. """
        return self._id

    @property
    def parent(self):
        """ The parent of this entity. """
        return self._parent

    @parent.setter
    def parent(self, parent):
        """ Set the parent of this entity to another entity. """

        # Make sure the parent is also an entity
        assert isinstance(parent, self.object.kind)
        self._parent = parent



from typing import Optional, Union
from enum import Enum
import numpy as np
import torch

tensor_or_ndarray = Union[torch.Tensor, np.ndarray]


class Point(Enum):
    """ Point feature, including x, y, z, intensity, depth, and mask. """
    INTENSITY = "intensity"
    DEPTH = "depth"
    MASK = "mask"
    X = "x"
    Y = "y"
    Z = "z"


class Pointcloud:
    """ A helper class that allow handling point cloud more easily with functionality
    like transforming point cloud and extracting features/properties from point cloud.
    Pointcloud can be built from either ``numpy.ndarray`` or ``torch.Tensor`` data.
    Methods will maintain the same data type.

    Args:
        xyz (tensor): x, y, z position of the point cloud. shape of ``(N,3)``
        intensity (tensor): Intensity of the point cloud. shape ``(N,)``

    """
    def __init__(self,
                 xyz: tensor_or_ndarray,
                 intensity: Optional[tensor_or_ndarray] = None):
        super(Pointcloud, self).__init__()

        self.with_torch = isinstance(xyz, torch.Tensor)
        self._xyz = xyz.reshape((-1, 3))
        self._intensity = 0.0 * self._xyz[:, 0]  # default to all zeros
        if intensity is not None:
            self._intensity[:] = intensity.ravel()

        self._dist: type(xyz) = None
        self._yaw: type(xyz) = None
        self._pitch: type(xyz) = None

    def transform(self,
                  R: Optional[tensor_or_ndarray] = None,
                  trans: Optional[tensor_or_ndarray] = None):
        """ Transform the point cloud.

        Args:
            R (tensor): Rotation matrix with shape (3,3).
            trans (tensor): Translation vector with length 3.

        Raises:
            AssertionError: Invalid rotation matrix (3,3) or translation (3,)

        """
        xyz = self.xyz

        if R is not None:
            assert R.shape == (3, 3), \
                   f"Rotation matrix shape is {R.shape} but should be (3, 3)"
            xyz = xyz @ R

        if trans is not None:
            assert trans.size == 3, \
                   f"trans size is {trans.size} but should be 3"
            trans = trans.reshape(1, 3)
            xyz = xyz + trans

        new_pcd = Pointcloud(xyz, self.intensity)
        return new_pcd

    def get(self, feature: Point) -> np.ndarray:
        """ Get feature (x, y, z, intensity, depth, mask) of the point cloud.

        Args:
            feature (Point): Feature to extract from the point cloud.

        Returns:
            np.ndarray: Point feature.

        Raises:
            ValueError: Unrecognized Point feature.

        """
        feature = Point(feature)  # Cast to a Point if not already
        if feature == Point.X:
            return self.x
        if feature == Point.Y:
            return self.y
        if feature == Point.Z:
            return self.z
        if feature == Point.INTENSITY:
            return self.intensity
        if feature == Point.DEPTH:
            return self.dist
        if feature == Point.MASK:
            ones = torch.ones if self.with_torch else np.ones
            return ones((self.num_points, ))

        raise ValueError(f"Unrecognized Point feature {feature} to" +
                         " extract from pointcloud")

    def __getitem__(self, i):
        new_pcd = Pointcloud(self.xyz[i], self.intensity[i])
        if self._dist is not None:
            new_pcd._dist = self._dist[i]
        return new_pcd

    def __len__(self):
        return self.xyz.shape[0]

    def numpy(self):
        """ Returns a copy of the torch pointcloud built using numpy. If the
        pointcloud is already in numpy format, then a copy is returned. """

        xyz = self.xyz
        intensity = self.intensity
        if self.with_torch:
            xyz = xyz.detach().cpu().numpy()
            intensity = intensity.detach().cpu().numpy()
        pcd_np = Pointcloud(xyz.copy(), intensity.copy())
        if self._dist is not None:
            dist = self._dist
            dist = dist.detach().cpu().numpy() if self.with_torch else dist
            pcd_np._dist = dist.copy()
        return pcd_np

    @property
    def num_points(self) -> int:
        """ Number of points. """
        return len(self)

    @property
    def x(self) -> tensor_or_ndarray:
        """ The `x` component of all points. """
        return self.xyz[:, 0]

    @property
    def y(self) -> tensor_or_ndarray:
        """ The `y` component of all points. """
        return self.xyz[:, 1]

    @property
    def z(self) -> tensor_or_ndarray:
        """ The `z` component of all points. """
        return self.xyz[:, 2]

    @property
    def xyz(self) -> tensor_or_ndarray:
        """ `xyz` of all points. """
        return self._xyz

    @property
    def intensity(self) -> tensor_or_ndarray:
        """ The intensity of all points. """
        return self._intensity

    @property
    def dist(self) -> tensor_or_ndarray:
        """ Distance to the origin of all points. """
        if self._dist is None:
            if self.with_torch:
                self._dist = torch.norm(self._xyz, p=2, dim=1)
            else:
                self._dist = np.linalg.norm(self._xyz, ord=2, axis=1)
        return self._dist

    @property
    def yaw(self) -> tensor_or_ndarray:
        """ Yaw angle (radians) of each point in the cloud. """
        if self._yaw is None:
            atan2 = torch.atan2 if self.with_torch else np.arctan2
            self._yaw = atan2(self.y, self.x)
        return self._yaw

    @property
    def pitch(self) -> tensor_or_ndarray:
        """ Pitch angle (radians) of each point in the cloud. """
        if self._pitch is None:
            arcsin = torch.arcsin if self.with_torch else np.arcsin
            self._pitch = arcsin(self.z / self.dist)
        return self._pitch

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} - ' + \
               f'#points: {self.num_points}' + '>'



from typing import Optional, List, Tuple
import numpy as np
import os
import xml.etree.ElementTree as ET
import pathlib
from ....utils.parse_params import ParamsFile


class CameraParams(object):
    """ The **CameraParams** object stores information pertaining to a single physical
    camera mounted on the car. It is useful for encapsulating the relevant calibration
    information for easy access in other modules.

    Args:
        rig_path (str): Path to RIG.xml that specifies camera parameters.
        name (str): Name of the camera identifier to initialize. Must be
                    a valid TopicName and present inside the RIG.xml file.
                    Can also specify `None` to auto grab the first named camera
                    in the RIG.xml file.
        params (dict): Dictionary camera parameters to instantiate with. If
                       not provided then the rig_path is used

    Raises:
        ValueError: if `name` is provided but not found in the rig file.

    """
    def __init__(self,
                 rig_path: str = None,
                 name: str = None,
                 params: dict = None):

        assert (rig_path is not None) ^ (params is not None)
        if rig_path is not None:
            pfile = ParamsFile(rig_path)
            params, self.name = pfile.parse_camera(name)

        self._height = int(params['height'])
        self._width = int(params['width'])

        self._fx = params['fx']
        self._fy = params['fy']

        self._cx = params['cx']
        self._cy = params['cy']

        self._distortion = params['distortion']
        self._quaternion = params['quaternion'].reshape(4, 1)
        self._position = params['position'].reshape(3, 1)
        self._yaw = params['yaw'] if 'yaw' in params else None

        self._roi = params['roi'].astype(np.int)
        self._roi_angle = params['roi_angle'] * np.pi / 180.

        self.__compute_other_forms()

    def resize(self, height: int, width: int) -> None:
        """ Scales the camera object and adjusts the internal parameters such
        that it projects images of a certain size.

        Args:
            height (int): New height of the camera images in pixels.
            width (int): New width of the camera images in pixels.

        """

        scale_y = float(height) / self._height
        scale_x = float(width) / self._width

        # Update height and width
        self._width = int(width)
        self._height = int(height)

        # Scale focal length
        self._fx *= scale_x
        self._fy *= scale_y

        # Scale optical center
        self._cx *= scale_x
        self._cy *= scale_y

        # ROI
        roi = [float(x) for x in self._roi]
        roi = [
            roi[0] * scale_y, roi[1] * scale_x, roi[2] * scale_y,
            roi[3] * scale_x
        ]
        self._roi = [int(x) for x in roi]

        self.__compute_other_forms()

    def crop(self, i1: int, j1: int, i2: int, j2: int) -> None:
        """ Crops a camera object to a given region of interest specified by
        the coordinates of the top left (i1,j1) and bottom right (i2,j2) corner.

        Args:
            i1 (int): Top row of ROI.
            j1 (int): Left column of ROI.
            i2 (int): Bottom row of ROI.
            j2 (int): Right column of ROI.

        """

        # Focal length stays the same
        self._fx = self._fx
        self._fy = self._fy

        # Height, width adjusted to ROI
        self._width = int(j2 - j1)
        self._height = int(i2 - i1)

        # Translate optical center to new frame of reference
        # NOTE: optical center (_cx,_cy) is measured from the bottom-right
        #   corner of the image.
        self._cx -= j1
        self._cy -= i2

        # ROI
        self._roi = [
            self._roi[0] - i1, self._roi[1] - j1, self._roi[2] - i2,
            self._roi[3] - j2
        ]
        self._roi = [int(x) for x in self._roi]

        self.__compute_other_forms()

    def get_height(self) -> int:
        """ Get the raw pixel height of images captured by the camera.

        Returns:
            int: Height in pixels.

        """
        return self._height

    def get_width(self) -> int:
        """ Get the raw pixel width of images captured by the camera.

        Returns:
            int: Width in pixels.

        """
        return self._width

    def get_K(self) -> np.ndarray:
        """ Get intrinsic calibration matrix.

        Returns:
            np.array: Intrinsic matrix (3,3).

        """
        return self._K

    def get_K_inv(self) -> np.ndarray:
        """ Get inverse intrinsic calibration matrix.

        Returns:
            np.array: Inverse intrinsic matrix (3,3).

        """
        return self._K_inv

    def get_distortion(self) -> np.ndarray:
        """ Get the distortion coefficients of the camera.

        Returns:
            np.array: Distortion coefficients (-1,).

        """
        return self._distortion

    def get_position(self) -> np.ndarray:
        """ Get the 3D position of camera.

        Returns:
            np.array: 3D position of camera.

        """
        return self._position

    def get_quaternion(self) -> np.ndarray:
        """ Get the rotation in quaternion of camera.

        Returns:
            np.array: Rotation in quaternion of camera.

        """
        return self._quaternion

    def get_yaw(self) -> float:
        """ Get the yaw of the camera relative the frame of reference.

        Returns:
            float: Yaw of the camera [rads].

        """
        if self._yaw is None:
            raise ValueError(
                f'camera {self.name}, does not have a yaw in the rig file')
        return self._yaw

    def get_ground_plane(self) -> List[float]:
        """ Get the equation of the ground plane.

        The equation of the ground plane is given by: ``Ax + By + Cz = D``
        and is computed from the position and orientation of the camera.

        Returns:
            List[float]: Parameterization of the ground plane: [A,B,C,D].

        """
        return self._ground_plane

    def get_roi(self, axis: Optional[str] = 'ij') -> List[int]:
        """ Get the region of interest of the images captured by the camera.

        Args:
            axis (str): Axis order to return the coordinates in (default 'ij',
            can also be 'xy').

        Returns:
            List[int]: Coordinates of the ROI box.

        Raises:
            ValueError: If `axis` is not valid.

        """

        if axis == 'ij':
            return self._roi
        elif axis == 'xy':
            return [self._roi[i] for i in [1, 2, 3, 0]]
        else:
            raise ValueError("invalid axis: " + axis)

    def get_roi_angle(self) -> float:
        """ Get the angle of the region of interest.

        Returns:
            float: The rotation of the ROI box.

        """
        return self._roi_angle

    def get_roi_points(self) -> List:
        """ Get the points of the region of interest.

        Returns:
            List: the list of points surrounding the ROI box

        """
        return self._roi_points

    def get_roi_dims(self) -> Tuple:
        """ Get the dimensions of the region of interest.

        Returns:
            Tuple: Height and width of ROI.

        """
        dims = (int(self._roi_height), int(self._roi_width))
        return dims if self._roi_angle < np.pi / 4. else dims[::-1]

    def __compute_other_forms(self):
        self.__compute_intrinsic_matrix()
        self.__compute_ground_plane()
        self.__compute_roi()

    def __compute_intrinsic_matrix(self):

        self._K = np.array([[self._fx, 0, self._cx], [0, self._fy, self._cy],
                            [0, 0, 1]])
        self._K_inv = np.linalg.inv(self._K)

    def __compute_ground_plane(self):
        q = self._quaternion.flatten().tolist()
        p = self._position

        normal = [
            2. * (q[1] * q[2] + q[3] * q[1]),
            1 - 2. * (q[0] * q[0] + q[2] * q[2]),
            2. * (q[0] * q[1] - q[2] * q[3])
        ]
        intercept = np.dot(p.T, normal)

        self._ground_plane = [normal[0], normal[1], normal[2], intercept[0]]

    def __compute_roi(self):
        t = self._roi_angle
        (i1, j1, i2, j2) = self._roi
        W = j2 - j1
        H = i2 - i1
        self._roi_width = (1. / (np.cos(t)**2 - np.sin(t)**2)) * (
            W * np.cos(t) - H * abs(np.sin(t)))
        self._roi_height = (1. / (np.cos(t)**2 - np.sin(t)**2)) * (
            -W * abs(np.sin(t)) + H * np.cos(t))
        if t < 0:  #reverse persective
            self._roi_points = [
                (abs(self._roi_height * np.cos(t)), 0),
                (H, abs(self._roi_width * np.cos(t))),
                (abs(self._roi_width * np.sin(t)), W),
                (0, abs(self._roi_height * np.sin(t))),
            ]
        else:
            self._roi_points = [
                (abs(self._roi_width * np.sin(t)), 0),
                (H, abs(self._roi_height * np.sin(t))),
                (abs(self._roi_height * np.cos(t)), W),
                (0, abs(self._roi_width * np.cos(t))),
            ]
        self._roi_points = [(i + i1, j + j1) for (i, j) in self._roi_points]
        self._roi_points = [(j, i) for (i, j) in self._roi_points]
        self._roi_points = np.int32([self._roi_points])



from typing import Optional, List, Tuple
import os
import numpy as np
import xml.etree.ElementTree as ET
import pathlib


def ignore_case(tree):
    t = ET.tostring(tree)
    t = t.lower()
    return ET.fromstring(t)


class ParamsFile:
    """ Helper object for processing a dataset parameters file. Every trace
    has a dedicated parameters file (``params.xml``) stored within it which
    contains calibration and other information on each of the sensors so we can
    reconstruct and simulate novel views.

    Args:
        path (str): Path to params.xml that specifies sensor parameters.

    """
    def __init__(self, path: str):
        self.path = path
        tree = ET.parse(self.path)
        self.root = ignore_case(tree.getroot())

    def parse_camera(self, name: str = None):
        """ Returns a parsed dictionary of all keys/values corresponding to the
        camera sensor with specified ``name``. If no ``name`` is given, then
        the first camera in the params file is used.

        Args:
            name (str): name of the camera (e.g. ``camera_front``). If not
            provided, then return the first camera.

        Returns:
            (tuple): tuple containing:
                (dict) of key -> value pairs for the specified camera sensor.
                (str): name of the sensor (useful if ``name`` was not specied
                as input)
        """
        return self._parse_sensor(sensor="camera", name=name)

    def parse_lidar(self, name: str = None):
        """ Returns a parsed dictionary of all keys/values corresponding to the
        lidar sensor with specified ``name``. If no ``name`` is given, then
        the first lidar in the params file is used.

        Args:
            name (str): name of the lidar (e.g. ``lidar_3d``). If not
            provided, then return the first lidar.

        Returns:
            (tuple): tuple containing:
                (dict) of key -> value pairs for the specified lidar sensor.
                (str): name of the sensor (useful if ``name`` was not specied
                as input)
        """
        return self._parse_sensor(sensor="lidar", name=name)

    def _parse_sensor(self, sensor: str, name: str = None):
        xml_sensors = self.root.findall(f'sensors/{sensor}')
        names = [cam.get('name') for cam in xml_sensors]
        sensors = dict(zip(names, xml_sensors))

        if name:
            if name not in sensors.keys():
                raise ValueError(
                    f'{name} not a valid name in {self.path}. ' + \
                    f'Only found sensors ({names})')
        else:
            # default to the first camera
            # TODO: should default to the closest camera (most overlapping FoV)
            name = names[0]

        self.name = name
        sensor = sensors[self.name]
        xml_props = sensor.findall('property')

        pname = [p.get('name') for p in xml_props]
        pvalue = []
        for p in xml_props:
            value = p.get('value')
            if "," in value:
                value = np.array([v for v in value.split(",")])
            else:
                value = np.array(value)

            try:
                value = value.astype(np.float32)
                if value.size == 1:
                    value = float(value)
            except ValueError:
                if value.size == 1:
                    value = str(value)
                else:
                    value = list(value)

            pvalue.append(value)

        props = dict(zip(pname, pvalue))
        return props, name



import imp
import importlib.resources as pkg_resources
import numpy as np
from scipy import interpolate
from typing import Tuple, Optional, Union
import torch
import warnings

import vista
from vista import resources
from vista.utils import transform, logging
from .Pointcloud import Pointcloud, Point
from .s2d_model import LidarModel

tensor_or_ndarray = Union[torch.Tensor, np.ndarray]


class LidarSynthesis:
    """ A Lidar synthesizer that simulates point cloud from novel viewpoint around
    a pre-collected Lidar sweep. At a high level, it involves (1) performing rigid
    transformation on point cloud based on given viewpoint change (2) projecting 3D
    point cloud to 2D image space with angle coordinates (3) densifying the sparse
    2D image (4) culling occluded region (5) masking out some points/pixels to simulate
    the sparse pattern of LiDAR sweep (6) reprojecting back to 3D point cloud or rays.

    Args:
        input_yaw_fov (float): Input LiDAR field of view in yaw axis; can be read from ``params.xml`` file.
        input_pitch_fov (float): Input LiDAR field of view in pitch axis; can be read from ``params.xml`` file.
        yaw_fov (float): Output LiDAR field of view in yaw axis; default is ``input_yaw_fov``.
        pitch_fov (float): Output LiDAR field of view in pitch axis; default is ``input_pitch_fov``.
        yaw_res (float): Resolution in yaw axis; default is ``0.1``.
        pitch_res (float): Resolution in pitch axis; default is ``0.1``.
        culling_r (int): The radius (from the origin) for culling occluded points.
        load_model (bool): Whether to load Lidar densifier model; default to ``True``.

    """
    def __init__(self,
                 input_yaw_fov: Tuple[float, float],
                 input_pitch_fov: Tuple[float, float],
                 yaw_fov: Optional[Tuple[float, float]] = None,
                 pitch_fov: Optional[Tuple[float, float]] = None,
                 yaw_res: float = 0.1,
                 pitch_res: float = 0.1,
                 culling_r: int = 1,
                 load_model: bool = True,
                 **kwargs):

        ### Basic properties required for setting up the synthesizer including
        # the dimensionality and resolution of the image representation space
        self._res = np.array([yaw_res, pitch_res], dtype=np.float32)
        self._fov = np.array([input_yaw_fov, input_pitch_fov],
                             dtype=np.float32)
        self._fov_rad = self._fov * np.pi / 180.

        self._dims = (self._fov[:, 1] - self._fov[:, 0]) / self._res
        self._dims = np.ceil(self._dims).astype(np.int)[:, np.newaxis]

        yaw_fov = input_yaw_fov if yaw_fov is None else yaw_fov
        pitch_fov = input_pitch_fov if pitch_fov is None else pitch_fov
        self._out_fov = np.array([yaw_fov, pitch_fov], dtype=np.float32)
        self._out_fov_rad = self._out_fov * np.pi / 180.

        ### Culling occluded LiDAR
        # Create a list of offset coordinates within a radius R of the origin,
        # but excluding the origin itself.
        self.device = 'cuda:0' if torch.cuda.is_available() else 'cpu:0'
        cull_axis = torch.arange(-culling_r, culling_r + 1)
        offsets = torch.meshgrid(cull_axis, cull_axis)
        offsets = torch.reshape(torch.stack(offsets, axis=-1), (-1, 2))  #(Nx2)
        offsets = offsets[torch.any(offsets != 0, axis=1)]  #remove origin
        offsets = offsets.to(self.device)
        self.offsets = offsets[None].type(torch.int32)  #expand_dims and cast

        ### Rendering masks and neural network model for sparse -> dense
        try:  # can only work with python 3.9
            rsrc_path = pkg_resources.files(resources)
        except AttributeError:
            with pkg_resources.path(vista, 'resources') as p:
                rsrc_path = p
        self.avg_mask = np.load(str(rsrc_path / "Lidar/avg_mask2.npy"))
        self.avg_mask_pt = torch.tensor(self.avg_mask).to(self.device)

        self.load_model = load_model
        self.render_model = None
        path = rsrc_path / "Lidar/LidarFillerDevens.pt"
        if path.is_file() and load_model:
            logging.debug(f"Loading Lidar model from {path}")

            state_dict = torch.load(path, map_location=self.device)
            self.render_model = LidarModel(layers=int(state_dict["layers"]),
                                           filters=int(state_dict["filters"]))
            self.render_model.load_state_dict(state_dict)
            self.render_model.to(self.device)
            self.render_model.eval()

    def synthesize(
        self,
        trans: np.ndarray,
        rot: np.ndarray,
        pcd: np.ndarray,
    ) -> Tuple[Pointcloud, np.ndarray]:
        """ Apply rigid transformation to a dense pointcloud and return new
        dense representation or sparse pointcloud.

        Args:
            trans (np.ndarray): Translation vector.
            rot (np.ndarray): Rotation matrix.
            pcd (np.ndarray): Point cloud.

        Returns:
            Returns a tuple (``pointcloud``, ``array``), where ``pointcloud``
            is the synthesized point cloud with view point change from the
            transform (``trans``, ``rot``), and ``array`` is the dense depth
            map in 2D image space.

        """
        # Rigid transform of points
        R = transform.rot2mat(rot)
        pcd = pcd.transform(R, trans)

        # Convert from new pointcloud to dense image
        sparse = self._pcd2sparse(pcd,
                                  channels=(Point.DEPTH, Point.INTENSITY,
                                            Point.MASK),
                                  return_as_tensor=True)

        # Find occlusions and cull them from the rendering
        occlusions = self._cull_occlusions(sparse[:, :, 0])
        sparse[occlusions[:, 0], occlusions[:, 1]] = np.nan

        # Densify the image before masking
        dense = self._sparse2dense(sparse, method="nn")

        # Sample the image to simulate active LiDAR using neural masking
        new_pcd = self._dense2pcd(dense)

        # Filter to the new pointcloud to match the desired output lidar specs
        new_pcd = new_pcd[(new_pcd.yaw > self._out_fov_rad[0, 0])
                          & (new_pcd.yaw < self._out_fov_rad[0, 1])]
        new_pcd = new_pcd[(new_pcd.pitch > self._out_fov_rad[1, 0])
                          & (new_pcd.pitch < self._out_fov_rad[1, 1])]
        # pitch = torch.arcsin(new_pcd.z / new_pcd.dist)
        # new_pcd = new_pcd[pitch < (14.0 / 180 * np.pi)]
        new_pcd = new_pcd.numpy()

        return (new_pcd, dense)

    def _pcd2sparse(self,
                    pcd: Pointcloud,
                    channels: Point = Point.DEPTH,
                    return_as_tensor: bool = False) -> tensor_or_ndarray:
        """ Convert from pointcloud to sparse image in polar coordinates.
        Fill image with specified features of the data (-1 = binary). """

        if not isinstance(channels, list) and not isinstance(channels, tuple):
            channels = [channels]

        # Compute the values to fill and the indicies where to fill
        values = [pcd.get(channel) for channel in channels]
        values = np.stack(values, axis=1)
        inds = self._compute_sparse_inds(pcd)

        # Re-order to fill points with smallest depth last
        order = np.argsort(pcd.dist)[::-1]
        values = values[order]
        inds = inds[:, order]

        # Creat the image and fill it
        img = np.empty((self._dims[1, 0], self._dims[0, 0], len(channels)),
                       np.float32)
        img.fill(np.nan)
        img[-inds[1], inds[0], :] = values
        return torch.tensor(img).to(self.device) if return_as_tensor else img

    def _cull_occlusions(
        self,
        sparse: tensor_or_ndarray,
        depth_slack: float = 0.1,
    ) -> tensor_or_ndarray:

        # Coordinates where we have depth samples
        coords = torch.stack(torch.where(sparse > 0)).T

        # At each location, also compute coordinate for all of its neighbors
        samples = coords[:, None, :] + self.offsets  # (N, M, 2)

        # Collect the samples in each neighborhood
        samples[:, :, 0] = torch.clip(samples[:, :, 0], 0, sparse.shape[0] - 1)
        samples[:, :, 1] = torch.clip(samples[:, :, 1], 0, sparse.shape[1] - 1)
        neighbor_depth = sparse[samples[:, :, 0], samples[:, :, 1]]

        # For each location, compute the average depth of all neighbors
        valid = ~torch.isnan(neighbor_depth)
        scalar_zero = torch.zeros(1, 1).to(self.device)
        neighbor_depth = torch.where(valid, neighbor_depth, scalar_zero)
        avg_depth = (torch.sum(neighbor_depth, axis=1) /
                     torch.sum(valid.to(torch.float), axis=1))

        # Estimate if the location is occluded by measuring if its depth
        # greater than its surroundings (i.e. if it is behind its surroundings)
        # Some amound of slack can be added here to allow for edge cases.
        my_depth = sparse[coords[:, 0], coords[:, 1]]
        occluded = (my_depth - depth_slack) > avg_depth

        # Return the coordinates in the depth image which are occluded and
        # should be disregarded
        occluded_coords = coords[occluded]
        return occluded_coords

    def _cull_occlusions_np(self,
                            sparse: np.ndarray,
                            depth_slack: float = 0.1) -> np.ndarray:

        coords = np.array(np.where(sparse > 0)).T

        samples = np.expand_dims(coords, 1) + self.offsets.numpy()
        samples[:, :, 0] = np.clip(samples[:, :, 0], 0, sparse.shape[0] - 1)
        samples[:, :, 1] = np.clip(samples[:, :, 1], 0, sparse.shape[1] - 1)

        neighbor_depth = sparse[samples[:, :, 0], samples[:, :, 1]]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            avg_depth = np.nanmean(neighbor_depth, axis=1)

        my_depth = sparse[coords[:, 0], coords[:, 1]]

        # point is valid if it is closer than the average depth around it
        occluded = (my_depth - depth_slack) > avg_depth
        occluded_coords = coords[occluded]

        # remove (cull) all invalid points
        # sparse[occluded_coords[:, 0], occluded_coords[:, 1]] = np.nan
        # return sparse
        return occluded_coords

    def _sparse2dense(self,
                      sparse: tensor_or_ndarray,
                      method: str = "linear") -> tensor_or_ndarray:
        """ Convert from sparse image representation of pointcloud to dense. """

        if method == "nn":
            sparse[torch.isnan(sparse)] = 0.0
            sparse = torch.nn.functional.pad(sparse, (0, 0, 4, 4, 4, 4))
            sparse = sparse[None].permute(0, 3, 1, 2)  # (1, 3, H+p, W+p)
            dense = self.render_model(sparse).detach()  # (1, 2, H+p, W+p)
            dense = dense.permute((0, 2, 3, 1))  # (1, H+p, W+p, 2)
            dense = dense[0, 4:-4, 4:-4]  # (H, W, 2)

        else:
            # mask all invalid values
            zs = np.ma.masked_invalid(sparse)

            # integer arrays for indexing
            grid_x, grid_y = np.meshgrid(np.arange(0, self._dims[0]),
                                         np.arange(0, self._dims[1]))

            # retrieve the valid, non-Nan, defined values
            valid_xs = grid_x[~zs.mask]
            valid_ys = grid_y[~zs.mask]
            valid_zs = zs[~zs.mask]

            # generate interpolated array of values
            dense = interpolate.griddata((valid_xs, valid_ys),
                                         valid_zs,
                                         tuple((grid_x, grid_y)),
                                         method=method)
            dense[np.isnan(dense)] = 0.0
        return dense

    def _dense2pcd(self, dense: tensor_or_ndarray):
        """ Sample mask from network and render points from mask """
        # TODO: load trained masking network and feed dense through
        # For now, simply load a mask prior from training data and sample
        mask_shape = self.avg_mask.shape
        if isinstance(dense, torch.Tensor):
            mask = self.avg_mask_pt > torch.rand(size=mask_shape).to(
                self.device)
            pitch, yaw = torch.where(mask)
        else:
            mask = self.avg_mask > np.random.uniform(size=mask_shape)
            pitch, yaw = np.where(mask)

        pitch, yaw = self._coords2angles(pitch, yaw)
        rays = self._angles2rays(pitch, yaw)

        sampled_depth_and_ints = dense[mask]
        dist = sampled_depth_and_ints[:, 0]

        intensity = None
        if dense.shape[-1] == 2:  # intensity dimension
            intensity = sampled_depth_and_ints[:, 1]

        points = (dist * rays).T
        pcd = Pointcloud(points, intensity)
        return pcd

    def _coords2angles(
        self, pitch_coords: tensor_or_ndarray, yaw_coords: tensor_or_ndarray
    ) -> Tuple[tensor_or_ndarray, tensor_or_ndarray]:

        yaw = yaw_coords * (self._fov_rad[0, 1] - self._fov_rad[0, 0]) / \
              self._dims[0, 0] + self._fov_rad[0, 0]
        pitch = pitch_coords * (self._fov_rad[1, 0] - self._fov_rad[1, 1]) / \
              self._dims[1, 0] + self._fov_rad[1, 1]
        return pitch, yaw

    def _angles2rays(self, pitch: tensor_or_ndarray,
                     yaw: tensor_or_ndarray) -> tensor_or_ndarray:

        with_torch = isinstance(pitch, torch.Tensor)
        cos = torch.cos if with_torch else np.cos
        sin = torch.sin if with_torch else np.sin
        stack = torch.stack if with_torch else np.array

        xyLen = cos(pitch)
        rays = stack([ \
            xyLen * cos(yaw),
            xyLen * sin(yaw),
            sin(pitch)])
        return rays

    def _compute_sparse_inds(self, pcd: Pointcloud) -> np.ndarray:
        """ Compute the indicies on the image representation which will be
        filled for a given pointcloud """

        # project point cloud to 2D point map
        yaw = np.arctan2(pcd.y, pcd.x)
        pitch = np.arcsin(pcd.z / pcd.dist)
        angles = np.stack((yaw, pitch))

        fov_range = self._fov_rad[:, [1]] - self._fov_rad[:, [0]]
        slope = self._dims / fov_range
        inds = slope * (angles - self._fov_rad[:, [0]])

        inds = np.floor(inds).astype(np.int)
        np.clip(inds, np.zeros((2, 1)), self._dims - 1, out=inds)

        return inds



from typing import Dict, Optional, Any

from ..Entity import Entity
from ...utils import misc


class BaseSensor(Entity):
    """ Base class of all sensors.

    Args:
        attach_to (Entity): A car to be attached to.
        config (dict): Configuration of the sensor.

    """
    DEFAULT_CONFIG = {}

    def __init__(self,
                 attach_to: Entity,
                 config: Optional[Dict] = None) -> None:
        super(BaseSensor, self).__init__()

        self._parent = attach_to
        config = misc.merge_dict(config, self.DEFAULT_CONFIG)
        self._config = config
        self._name = config['name']

    def capture(self, timestamp: float, **kwargs) -> Any:
        """ Run sensor synthesis based on current timestamp and transformation
        between the novel viewpoint to be simulated and the nominal viewpoint from
        the pre-collected dataset.

        Args:
            timestamp (float): Timestamp that allows to retrieve a pointer to
                the dataset for data-driven simulation (synthesizing point cloud
                from real LiDAR sweep).

        """
        raise NotImplementedError

    def update_scene_object(self, name: str, scene_object: Any,
                            pose: Any) -> None:
        """ Update object in the scene for rendering. This is only used
        when we put virtual objects in the scene.

        Args:
            name (str): Name of the scene object.
            scene_object (Any): The scene object.
            pose (Any): The pose of the scene object.

        """
        raise NotImplementedError

    @property
    def name(self) -> str:
        """ The name of the sensor. """
        return self._name



class TopicNames:
    """ This class specifies (ROS) topic names of each sensors.
    """
    camera_front = 'camera_front'
    camera_wing_left = 'camera_wing_left'
    camera_wing_right = 'camera_wing_right'
    camera_left = 'camera_left'
    camera_right = 'camera_right'
    cameras = [
        camera_front, camera_wing_right, camera_wing_right, camera_left,
        camera_right
    ]

    steering = 'steering_can'
    imu = 'imu'
    odometry = 'odom'
    gps = 'gps'
    gps_heading = 'gps_heading'
    speed = 'speed'
    distance = 'distance'
    lidar_3d = 'lidar_3d'

    BADVAL = ['BADVAL']  # for denoting topics we don't know yet

    # Master sensor to align all others with:
    master_topic = camera_front



import os
from typing import Dict, Any
import numpy as np
import h5py

from .BaseSensor import BaseSensor
from .lidar_utils import LidarSynthesis, Pointcloud
from ..Entity import Entity
from ...core import Trace
from ...utils import logging, parse_params


class Lidar(BaseSensor):
    """ A LiDAR sensor object that synthesizes LiDAR measurement locally around the
    dataset given a viewpoint (potentially different from the dataset) and timestamp.

    Args:
        attach_to (Entity): A car to be attached to.
        config (dict): Configuration of LiDAR sensor. An example (default) is,

            >>> DEFAULT_CONFIG = {
                'name': 'lidar_3d',
                'yaw_fov': None,
                'pitch_fov': None,
                'culling_r': 1,
                'use_synthesizer': True,
            }

            Check :class:`Lidarsynthesis` object for more details about the configuration.

    """
    DEFAULT_CONFIG = {
        'name': 'lidar_3d',
        'yaw_fov': None,
        'pitch_fov': None,
        'culling_r': 1,
        'use_synthesizer': True,
    }

    def __init__(self, attach_to: Entity, config: Dict) -> None:
        super(Lidar, self).__init__(attach_to, config)

        logging.debug('Not actually streaming lidar data when reading')
        self._streams: Dict[str, h5py.File] = dict()

        # Initialize lidar novel view synthesis object
        if self.config['use_synthesizer']:
            # Make one view synthesizer for each trace within
            # attach_to.parent.traces (different traces may have different
            # input sensors specs). For each synthesizer, pass in the input
            # and output lidar params. Then during synthesis, use the
            # appropraite synthesizer.
            self._view_synthesizers = {}
            for trace in attach_to.parent.traces:
                pfile = parse_params.ParamsFile(trace.param_file)
                in_params, _ = pfile.parse_lidar(self.name)
                self._view_synthesizers[trace] = LidarSynthesis(
                    load_model=True,
                    input_yaw_fov=in_params['yaw_fov'],
                    input_pitch_fov=in_params['pitch_fov'],
                    **self.config)

        else:
            self._view_synthesizers = None

    def reset(self) -> None:
        """ Reset LiDAR sensor by initiating LiDAR data stream based on
        current reference pointer to the dataset.

        """
        logging.info(f'Lidar ({self.id}) reset')

        # Initiate lidar data stream based on current reference pointer to the
        # dataset. All data streams are handled by the main lidar and shared
        # across all Lidar objects in an agent.
        multi_sensor = self.parent.trace.multi_sensor
        if self.name == multi_sensor.main_lidar:
            for lidar_name in multi_sensor.lidar_names:
                fpath = os.path.join(self.parent.trace.trace_path,
                                     lidar_name + '.h5')
                stream = h5py.File(fpath, 'r')
                self._streams[lidar_name] = stream
        else:
            main_name = multi_sensor.main_lidar
            main_sensor = [
                _s for _s in self.parent.sensors if _s.name == main_name
            ]
            assert len(main_sensor) == 1, \
                    f'Cannot find main sensor {main_name}'

            main_sensor = main_sensor[0]
            assert isinstance(main_sensor, Lidar), \
                    'Main sensor is not Lidar object'
            self._streams = main_sensor.streams

        # reset lidar synthesis
        pass

    def capture(self, timestamp: float, **kwargs) -> np.ndarray:
        """ Synthesize LiDAR point cloud based on current timestamp and transformation
        between the novel viewpoint to be simulated and the nominal viewpoint from the
        pre-collected dataset.

        Args:
            timestamp (float): Timestamp that allows to retrieve a pointer to
                the dataset for data-driven simulation (synthesizing point cloud
                from real LiDAR sweep).

        Returns:
            np.ndarray: Synthesized point cloud.

        """
        logging.info(f'Lidar ({self.id}) capture')

        # Get frame at the closest smaller timestamp from dataset.
        multi_sensor = self.parent.trace.multi_sensor
        all_frame_nums = multi_sensor.get_frames_from_times([timestamp])
        for lidar_name in multi_sensor.lidar_names:
            stream = self.streams[lidar_name]
            frame_num = all_frame_nums[lidar_name][0]
            xyz = stream['xyz'][frame_num]
            intensity = stream['intensity'][frame_num]
            pcd = Pointcloud(xyz, intensity)
            pcd = pcd[pcd.dist > 2.5]
            # TODO: when is it possible for there to be multiple (multi_sensor.lidar_names)?

        # Synthesis by rendering
        if self.config['use_synthesizer']:
            # self.parent.reslative_state.update(0, 0, yaw=np.sin(timestamp))
            lat, long, yaw = self.parent.relative_state.numpy()
            logging.debug(
                f"state: {lat} {long} {yaw} \t timestamp {timestamp}")
            trans = np.array([-long, lat, 0])
            rot = np.array([0., 0, yaw])  # TODO: should yaw be Y or Z?
            synthesizer = self._view_synthesizers[self.parent.trace]
            new_pcd, new_dense = synthesizer.synthesize(trans, rot, pcd)
        else:
            new_pcd = pcd

        return new_pcd

    def update_scene_object(self, name: str, scene_object: Any,
                            pose: Any) -> None:
        """ Adding virtual object in LiDAR synthesis is not yet implemented. """
        raise NotImplementedError

    @property
    def config(self) -> Dict:
        """ Configuration of the LiDAR sensor. """
        return self._config

    @property
    def streams(self) -> Dict[str, h5py.File]:
        """ Data stream of LiDAR dataset to be simulated from. """
        return self._streams

    @property
    def view_synthesis(self) -> LidarSynthesis:
        """ Wiew synthesizer object for the first trace. """
        first = list(self._view_synthesizers.keys())[0]
        return self._view_synthesizers[first]



from typing import Optional, List
import numpy as np
import scipy.integrate as ode_solve

from ...utils import logging


class State:
    """ Vehicle state without steering angle and velocity. """
    def __init__(self,
                 x: Optional[float] = 0.,
                 y: Optional[float] = 0.,
                 yaw: Optional[float] = 0.) -> None:
        self.update(x, y, yaw)

    def update(self, x: float, y: float, yaw: float) -> None:
        """ Set values. """
        self._x = x
        self._y = y
        self._yaw = yaw

    def reset(self) -> None:
        """ Reset to zeros. """
        self.update(0., 0., 0.)

    def numpy(self) -> np.ndarray:
        """ Convert to numpy. """
        return np.array([self._x, self._y, self._yaw])

    @property
    def x(self) -> float:
        """ Position of the car in x-axis. """
        return self._x

    @property
    def y(self) -> float:
        """ Position of the car in y-axis. """
        return self._y

    @property
    def yaw(self) -> float:
        """ Heading/yaw of the car. """
        return self._yaw

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: ' + \
               f'[{self.x}, {self.y}, {self.yaw}]>'


class StateDynamics:
    """ Simple continuous kinematic (bicycle) model of a rear-wheel driven vehicle.
    Check out Eq.3 in https://www.autonomousrobots.nl/docs/17-schwartig-autonomy-icra.pdf

    Args:
        x (float): Position of the car in x-axis.
        y (float): Position of the car in y-axis.
        yaw (float): Heading/yaw of the car.
        steering (float): Steering angle of tires instead of steering wheel.
        speed (float): Forward speed.
        steering_bound (list): Upper and lower bound of steering angle.
        speed_bound (list): Upper and lower bound of speed.
        wheel_base(float): Wheel base.

    """
    def __init__(self,
                 x: Optional[float] = 0.,
                 y: Optional[float] = 0.,
                 yaw: Optional[float] = 0.,
                 steering: Optional[float] = 0.,
                 speed: Optional[float] = 0.,
                 steering_bound: Optional[List[float]] = [-0.75, 0.75],
                 speed_bound: Optional[List[float]] = [0., 15.],
                 wheel_base: Optional[float] = 2.8) -> None:
        self._x = 0.
        self._y = 0.
        self._yaw = 0.
        self._steering = 0.
        self._speed = 0.

        self.update(x, y, yaw, steering, speed)
        self._steering_bound = steering_bound
        self._speed_bound = speed_bound
        self._wheel_base = wheel_base

    def step(self,
             steering_velocity: float,
             acceleration: float,
             dt: float,
             max_steps: Optional[int] = 100) -> np.ndarray:
        """ Step the vehicle state by solving ODE of the continuous kinematic (bicycle) model.

        Args:
            steering_velocity (float): Steering velocity (of tire angle).
            acceleration (float): Acceleration.
            dt (float): Elapsed time.
            max_steps (int): Maximal step to run the ODE solver; default to 100.

        Returns:
            np.ndarray: The updated state (x, y, yaw, tire angle, speed).

        """

        # Define dynamics
        def _ode_func(t, z):
            _x, _y, _phi, _delta, _v = z
            u_delta = steering_velocity
            u_a = acceleration
            new_z = np.array([
                -_v * np.sin(_phi),  # swap x-y axis with sign change
                _v * np.cos(_phi),
                _v / self._wheel_base * np.tan(_delta),
                u_delta,
                u_a
            ])
            return new_z

        # Solve ODE
        z_0 = np.array(
            [self._x, self._y, self._yaw, self._steering, self._speed])
        solver = ode_solve.RK45(_ode_func, 0., z_0, dt)
        steps = 0
        while solver.status is 'running' and steps <= max_steps:
            solver.step()
            steps += 1
        if (dt - solver.t) < 0:
            logging.error('Reach max steps {} without reaching t_bound ({} < {})'.format( \
                max_steps, solver.t, solver.t_bound))

        self._x, self._y, self._yaw, self._steering, self._speed = solver.y

        # Clip by value bounds
        self._steering = np.clip(self._steering, *self._steering_bound)
        self._speed = np.clip(self._speed, *self._speed_bound)

        return self.numpy()

    def numpy(self) -> np.ndarray:
        """ Return a numpy array of vehicle state.

        Returns:
            np.ndarray: Vehicle state.

        """
        return np.array(
            [self._x, self._y, self._yaw, self._steering, self._speed])

    def copy(self):
        """ Create a copy. """
        return StateDynamics(x=self._x,
                             y=self._y,
                             yaw=self._yaw,
                             steering=self._steering,
                             speed=self._speed)

    def update(self, x: float, y: float, yaw: float, steering: float,
               speed: float) -> None:
        """ Set state.
        Args:
            x (float): Position of the car in x-axis.
            y (float): Position of the car in y-axis.
            yaw (float): Heading/yaw of the car.
            steering (float): Steering angle of tires instead of steering wheel.
            speed (float): Forward speed.

        """
        self._x = x
        self._y = y
        self._yaw = yaw
        self._steering = steering
        self._speed = speed

    def reset(self) -> None:
        """ Reset all state to zeros. """
        self.update(0., 0., 0., 0., 0.)

    @property
    def x(self) -> float:
        """ Position of the car in x-axis. """
        return self._x

    @property
    def y(self) -> float:
        """ Position of the car in y-axis. """
        return self._y

    @property
    def yaw(self) -> float:
        """ Heading/yaw of the car. """
        return self._yaw

    @property
    def steering(self) -> float:
        """ Steering (tire) angle. """
        return self._steering

    @property
    def steering_bound(self) -> List[float]:
        """ Lower and upper bound of steering (tire) angle. """
        return self._steering_bound

    @property
    def speed(self) -> float:
        """ Speed. """
        return self._speed

    @property
    def speed_bound(self) -> List[float]:
        """ Lower and upper bound of speed. """
        return self._speed_bound

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: ' + \
            f'[{self.x}, {self.y}, {self.y}, {self.steering}, {self.speed}]>'


def curvature2tireangle(curvature: float, wheel_base: float) -> float:
    """ Convert curvature to tire angle.

    Args:
        curvature (float): Curvature.
        wheel_base (float): Wheel base.

    Returns:
        float: Tire angle.

    """
    return np.arctan(wheel_base * curvature)


def tireangle2curvature(tire_angle: float, wheel_base: float) -> float:
    """ Convert tire angel to curvature.

    Args:
        tire_angle (float): Tire angle.
        wheel_base (float): Wheel base.

    Returns:
        float: Curvature.

    """
    return np.tan(tire_angle) / wheel_base


def curvature2steering(curvature: float, wheel_base: float,
                       steering_ratio: float) -> float:
    """ Convert curvature to steering angle.

    Args:
        curvature (float): Curvature.
        wheel_base (float): Wheel base.
        steering_ratio (float): Steering ratio.

    Returns:
        float: Steering (wheel) angle.

    """
    tire_angle = curvature2tireangle(curvature, wheel_base)
    steering = tire_angle * steering_ratio * 180. / np.pi

    return steering


def steering2curvature(steering: float, wheel_base: float,
                       steering_ratio: float) -> float:
    """ Convert steering angle to curvature.

    Args:
        steering (float): Steering (wheel) angle.
        wheel_base (float): Wheel base.
        steering_ratio (float): Steering ratio.

    Returns:
        float: Curvature.

    """
    tire_angle = steering * (np.pi / 180.) / steering_ratio
    curvature = tireangle2curvature(tire_angle, wheel_base)

    return curvature


def update_with_perfect_controller(desired_state: List[float], dt: float,
                                   dynamics: StateDynamics) -> None:
    """ Update vehicle state assuming a perfect low-level controller. This
    basically simulate that the low-level controller can "instantaneously"
    achieve the desired state (e.g., steering (tire) angle and speed) with control
    command (steering velocity and accleration) in vehicle dynamics.

    Args:
        desired_state (List[float]): A list of desired vehicle state.
        dt (float): Elapsed time.
        dynamics (StateDynamics): Vehicle state.

    """
    # simulate condition when the desired state can be instantaneously achieved
    new_dyn = dynamics.numpy()
    new_dyn[-2:] = desired_state
    dynamics.update(*new_dyn)
    dynamics.step(0., 0., dt)



from typing import Optional, Dict, Union, List, Tuple, Any
import numpy as np
import cv2
from shapely.geometry import box as Box
from shapely import affinity

from . import transform
from ..entities.agents import Car
from ..entities.agents.Dynamics import StateDynamics

Vec = Union[List, Tuple, np.ndarray]


def agent2poly(agent: Car,
               ref_dynamics: Optional[StateDynamics] = None) -> Box:
    """ Convert Agent object to polygon w.r.t. a reference dynamics.

    Args:
        agent (Car): An agent with valid dynamics (pose and vehicle state).
        ref_dynamics (StateDynamics):
            A reference dynamics for computing the polygon representation
            of the agent. Default set to None that uses human dynamics of
            the agent.

    Returns:
        Box: A polygon that describes the agent.

    """
    ref_dynamics = agent.human_dynamics if ref_dynamics is None else ref_dynamics
    rel_pose = transform.compute_relative_latlongyaw(
        agent.ego_dynamics.numpy()[:3],
        ref_dynamics.numpy()[:3])
    poly = Box(rel_pose[0] - agent.width / 2., rel_pose[1] - agent.length / 2.,
               rel_pose[0] + agent.width / 2., rel_pose[1] + agent.length / 2.)
    poly = affinity.rotate(poly, np.degrees(rel_pose[2]))
    return poly


def merge_dict(dict1: Dict, dict2: Dict) -> Dict:
    """ Merge two dict, where dict1 has higher priority.

    Args:
        dict1 (Dict): The first dictionary.
        dict2 (Dict): The second dictionary.

    Returns:
        Dict: The merged dictionary.

    """
    return dict(list(dict2.items()) + list(dict1.items()))


def fetch_agent_info(agent: Car) -> Dict[str, Any]:
    """ Get info from agent class.

    Args:
        agent (Car): The agent to be extracted information from.

    Returns:
        Dict: A dictionary contains various information of the agent.

    """
    info = dict(
        trace_path=agent.trace.trace_path,
        relative_state=agent.relative_state.numpy(),
        ego_dynamics=agent.ego_dynamics.numpy(),
        human_dynamics=agent.human_dynamics.numpy(),
        length=agent.length,
        width=agent.width,
        wheel_base=agent.wheel_base,
        steering_ratio=agent.steering_ratio,
        speed=agent.speed,
        curvature=agent.curvature,
        steering=agent.steering,
        tire_angle=agent.tire_angle,
        human_speed=agent.human_speed,
        human_curvature=agent.human_curvature,
        human_steering=agent.human_steering,
        human_tire_angle=agent.human_tire_angle,
        timestamp=agent.timestamp,
        frame_number=agent.frame_number,
        trace_index=agent.trace_index,
        segment_index=agent.segment_index,
        frame_index=agent.frame_index,
        trace_done=agent.done,
    )
    return info


def img2flow(img: np.ndarray,
             mag_minmax: Vec,
             flow_size: Optional[Vec] = None) -> np.ndarray:
    """ Convert HSV-encoded flow image to optical flow.

    Args:
        img (np.ndarray): An image with channel order BGR.
        mag_minmax (Vec): The minmum and maximum when normalizing the
                          flow magnitude to [0,1].
        flow_size (Vec): Size of the output flow array. If set, resize
                         the image before converting to flow; default
                         to ``None``.

    Returns:
        np.ndarray:
            A HxWx2 array with the two channels as magnitude and the angle of the flow.

    """
    hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)  # assume image is BGR
    if flow_size is not None:
        hsv = cv2.resize(hsv, flow_size[::-1])
    ang = hsv[..., 0] * 2. * np.pi / 180
    mag = hsv[..., 1] / 255. * (mag_minmax[1] - mag_minmax[0]) + mag_minmax[0]
    flow = np.stack(cv2.polarToCart(mag, ang), axis=-1)
    return flow


def biinterp(I0: np.ndarray, I1: np.ndarray, F_0_1: np.ndarray,
             F_1_0: np.ndarray, ts: float, t0: float, t1: float) -> np.ndarray:
    """ Interpolate frame with bidirectional flow.

    Args:
        I0 (np.ndarray): A RGB image at time `t0`.
        I1 (np.ndarray): A RGB image at time `t1`.
        F_0_1 (np.ndarray): The flow from time `t0` to `t1`.
        F_1_0 (np.ndarray): The flow from time `t1` to `t0`.
        ts (float): The timestamp to be interpolated to.
        t0 (float): The timestamp of `I0`.
        t1 (float): The timestamp of `I1`.

    Returns:
        np.ndarray: An interpolated RGB image at time `ts`.

    """
    t = (ts - t0) / (t1 - t0)
    temp = -t * (1 - t)
    fCoeff = [temp, t * t, (1 - t) * (1 - t), temp]
    F_t_0 = fCoeff[0] * F_0_1 + fCoeff[1] * F_1_0
    F_t_1 = fCoeff[2] * F_0_1 + fCoeff[3] * F_1_0

    g_I0_F_t_0 = flow_backwarp(I0, F_t_0)
    g_I1_F_t_1 = flow_backwarp(I1, F_t_1)

    wCoeff = [1 - t, t]
    out = wCoeff[0] * g_I0_F_t_0 + wCoeff[1] * g_I1_F_t_1
    return out


def flow_backwarp(img: np.ndarray,
                  flow: np.ndarray,
                  use_pytorch: Optional[bool] = False) -> np.ndarray:
    """ Warp image based on optical flow.

    Args:
        img (np.ndarray): An image to be warped.
        flow (np.ndarray): Optical flow to warp the image.
        use_pytorch (bool): Whether to use pytorch for warping; default to ``False``.

    Returns:
        np.ndarray: A warped image.

    """
    H, W = img.shape[:2]
    gridX, gridY = np.meshgrid(np.arange(W), np.arange(H))

    u = flow[:, :, 0]
    v = flow[:, :, 1]
    x = gridX + u
    y = gridY + v

    if use_pytorch:
        import torch
        x = 2 * (x / W - 0.5)
        y = 2 * (y / H - 0.5)
        grid = torch.from_numpy(np.stack((x, y), axis=2))

        img = torch.from_numpy(img / 255.).permute(2, 0, 1)
        out = torch.nn.functional.grid_sample(img[None, ...],
                                              grid[None, ...],
                                              align_corners=True)
        out = (out[0].permute(1, 2, 0).cpu().numpy() * 255.).astype(np.uint8)
    else:
        img = img / 255.
        grid = np.stack([x, y], axis=-1)
        out = cv2.remap(img, x.astype(np.float32), y.astype(np.float32),
                        cv2.INTER_LINEAR)
        out = np.clip(out * 255, 0, 255).astype(np.uint8)
    return out



from threading import Event
from typing import Optional, Dict, Tuple, List, Any
from collections import deque
import cv2
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm, colors, patches

from shapely.geometry import LineString
from descartes import PolygonPatch
import cv2

from . import World
from ..entities.agents.Dynamics import StateDynamics, update_with_perfect_controller, \
                                       curvature2tireangle
from ..entities.agents import Car
from ..entities.sensors import Camera, EventCamera, Lidar
from ..entities.sensors.camera_utils import CameraParams
from ..entities.sensors.lidar_utils import Pointcloud
from ..utils import logging, transform, misc


class Display:
    """ This is a visualizer of VISTA simulator. It renders an image that contains visualization
    of all sensors from all agents and a top-down view that depicts road and all cars in the scene
    within a predefined range based on the state of the simulator (:class:`World`).

    Args:
        world (vista.core.World): World to be visualized.
        fps (int): Frame per second.
        display_config (Dict): Configuration for the display (visualization).

    Raises:
        AssertionError: Grid spec is inconsistent with maximal number of sensors across agents.

    Example usage::

        >>> display_config = {
            'road_buffer_size': 200,
            'birdseye_map_size': (30, 20), # size of bev map in vertical and horizontal directions
            'gs_bev_w': 2, # grid spec width for the birdseye view block
            'gs_agent_w': 4, # grid spec width for an agent's block
            'gs_h': 6, # grid spec height
            'gui_scale': 1.0, # a global scale that determines the size of the figure
            'vis_full_frame': False, # if Display should not crop/resize camera for visualization purposes
        }
        >>> display = Display(world, )

    """
    DEFAULT_DISPLAY_CONFIG = {
        'road_buffer_size': 200,
        'birdseye_map_size': (30, 20),
        'gs_bev_w': 2,
        'gs_agent_w': 4,
        'gs_h': 6,
        'gui_scale': 1.0,
        'vis_full_frame': False
    }

    def __init__(self,
                 world: World,
                 fps: Optional[float] = 30,
                 display_config: Optional[Dict] = dict()):
        # Get arguments
        self._world: World = world
        self._fps: float = fps
        self._config: Dict = misc.merge_dict(display_config,
                                             self.DEFAULT_DISPLAY_CONFIG)

        # Initialize data for plotting
        self._road: deque[np.ndarray] = deque(
            maxlen=self._config['road_buffer_size'])
        self._road_frame_idcs: deque[int] = deque(
            maxlen=self._config['road_buffer_size'])
        self._road_dynamics: StateDynamics = StateDynamics()

        # Get agents with sensors (to be visualized)
        self._agents_with_sensors: List[Car] = []
        for agent in self._world.agents:
            if len(agent.sensors) > 0:
                is_camera = [isinstance(_v, Camera) for _v in agent.sensors]
                if len(is_camera) > 0:
                    self._agents_with_sensors.append(agent)
                if len(is_camera) < len(agent.sensors):
                    logging.warning(
                        'Cannot visualize sensor other than Camera')
        n_agents_with_sensors = len(self._agents_with_sensors)
        if n_agents_with_sensors > 0:
            max_n_sensors = max(
                [len(_v.sensors) for _v in self._agents_with_sensors])
        else:
            max_n_sensors = 1

        # Specify colors for agents and road
        colors = list(cm.get_cmap('Set1').colors)
        rgba2rgb = lambda rgba: np.clip(
            (1 - rgba[:3]) * rgba[3] + rgba[:3], 0., 1.)
        colors = [np.array(list(c) + [0.6]) for c in colors]
        colors = list(map(rgba2rgb, colors))
        self._agent_colors: List[Tuple] = colors
        self._road_color: Tuple = list(cm.get_cmap('Dark2').colors)[-1]

        # Initialize figure
        self._artists: Dict[Any] = dict()
        self._axes: Dict[plt.Axes] = dict()
        gui_scale = self._config['gui_scale']
        figsize = (6.4 * gui_scale * n_agents_with_sensors + 3.2 * gui_scale,
                   3.2 * gui_scale * max_n_sensors)
        self._fig: plt.Figure = plt.figure(figsize=figsize)
        self._fig.patch.set_facecolor('black')  # use black background
        self._gs = self._fig.add_gridspec(
            self._config['gs_h'],
            self._config['gs_agent_w'] * n_agents_with_sensors +
            self._config['gs_bev_w'])
        assert self._config['gs_h'] % max_n_sensors == 0, \
            (f'Height of grid ({self._config["gs_h"]}) can not be exactly ' + \
            f'divided by max number of sensors ({max_n_sensors})')
        gs_agent_h = self._config['gs_h'] // max_n_sensors

        # Initialize birds eye view
        self._axes['bev'] = self._fig.add_subplot(
            self._gs[:, -self._config['gs_bev_w']:])
        self._axes['bev'].set_facecolor('black')
        self._axes['bev'].set_xticks([])
        self._axes['bev'].set_yticks([])
        self._axes['bev'].set_title('Top-down View',
                                    color='white',
                                    size=20,
                                    weight='bold')
        self._axes['bev'].set_xlim(-self._config['birdseye_map_size'][1] / 2.,
                                   self._config['birdseye_map_size'][1] / 2.)
        self._axes['bev'].set_ylim(-self._config['birdseye_map_size'][0] / 2.,
                                   self._config['birdseye_map_size'][0] / 2.)

        # Initialize plot for sensory measurement
        logging.debug(
            'Does not handle preprocessed (cropped/resized) observation')
        for i, agent in enumerate(self._agents_with_sensors):
            for j, sensor in enumerate(agent.sensors):

                if (isinstance(sensor, Camera)
                        or isinstance(sensor, EventCamera)):
                    param = sensor.camera_param
                    img_shape = (param.get_height(), param.get_width(), 3)

                elif isinstance(sensor, Lidar):
                    x_dim, y_dim = sensor.view_synthesis._dims[:, 0]
                    # Cut width in half and stack on-top
                    img_shape = (y_dim * 2, x_dim // 2, 3)

                else:
                    logging.error(f'Unrecognized sensor type {type(sensor)}')
                    continue

                gs_ij = self._gs[gs_agent_h * j:gs_agent_h * (j + 1),
                                 self._config['gs_agent_w'] *
                                 i:self._config['gs_agent_w'] * (i + 1)]
                ax_name = 'a{}s{}'.format(i, j)
                self._axes[ax_name] = self._fig.add_subplot(gs_ij,
                                                            facecolor='black')
                self._axes[ax_name].set_xticks([])
                self._axes[ax_name].set_yticks([])
                self._axes[ax_name].set_title('Init',
                                              color='white',
                                              size=20,
                                              weight='bold')

                placeholder = fit_img_to_ax(
                    self._fig, self._axes[ax_name],
                    np.zeros(img_shape, dtype=np.uint8))
                self._artists['im:{}'.format(
                    ax_name)] = self._axes[ax_name].imshow(placeholder)

        self._fig.tight_layout()

    def reset(self) -> None:
        """ Reset the visualizer. This should be called every time after
        :class:`World` reset. It basically reset the cache of road data
        used in the top-down view visualization.
        """
        # Reset road deque
        self._road.clear()
        self._road.append(self.ref_agent.human_dynamics.numpy()[:3])
        self._road_dynamics = self.ref_agent.human_dynamics.copy()
        self._road_frame_idcs.clear()
        self._road_frame_idcs.append(self.ref_agent.frame_index)

    def render(self):
        """ Render an image that visualizes the simulator. This includes visualization
        of all sensors of every agent and a top-down view that depicts the road and all
        cars in the scene within a certain range. Note that it render visualization based
        on the current status of the world and should be called every time when there is
        any update to the simulator.

        Returns:
            np.ndarray: An image of visualization for the simulator.

        """
        # Update road (in global coordinate)
        exceed_end = False
        while self._road_frame_idcs[-1] < (
                self.ref_agent.frame_index +
                self._config['road_buffer_size'] / 2.) and not exceed_end:
            exceed_end, ts = self._get_timestamp(self._road_frame_idcs[-1])
            self._road_frame_idcs.append(self._road_frame_idcs[-1] + 1)
            exceed_end, next_ts = self._get_timestamp(
                self._road_frame_idcs[-1])

            state = [
                curvature2tireangle(self.ref_agent.trace.f_curvature(ts),
                                    self.ref_agent.wheel_base),
                self.ref_agent.trace.f_speed(ts)
            ]
            update_with_perfect_controller(state, next_ts - ts,
                                           self._road_dynamics)
            self._road.append(self._road_dynamics.numpy()[:3])

        # Update road in birds eye view (in reference agent's coordinate)
        ref_pose = self.ref_agent.human_dynamics.numpy()[:3]
        logging.debug('Computation of road in reference frame not vectorized')
        road_in_ref = np.array([
            transform.compute_relative_latlongyaw(_v, ref_pose)
            for _v in self._road
        ])
        road_half_width = self.ref_agent.trace.road_width / 2.
        patch = LineString(road_in_ref).buffer(road_half_width)
        patch = PolygonPatch(patch,
                             fc=self._road_color,
                             ec=self._road_color,
                             zorder=1)
        self._update_patch(self._axes['bev'], 'patch:road', patch)

        # Update agent in birds eye view
        for i, agent in enumerate(self._world.agents):
            poly = misc.agent2poly(agent, self.ref_agent.human_dynamics)
            color = self._agent_colors[i]
            patch = PolygonPatch(poly, fc=color, ec=color, zorder=2)
            self._update_patch(self._axes['bev'], 'patch:agent_{}'.format(i),
                               patch)

        # Update sensory measurements
        for i, agent in enumerate(self._agents_with_sensors):
            cameras = {
                _v.name: _v
                for _v in agent.sensors if isinstance(_v, Camera)
            }
            event_cameras = {
                _v.name: _v
                for _v in agent.sensors if isinstance(_v, EventCamera)
            }
            lidars = {
                _v.name: _v
                for _v in agent.sensors if isinstance(_v, Lidar)
            }
            for j, (obs_name, obs) in enumerate(agent.observations.items()):
                ax_name = 'a{}s{}'.format(i, j)
                if obs_name in cameras.keys():
                    obs = plot_roi(obs.copy(),
                                   cameras[obs_name].camera_param.get_roi())
                    sensor = cameras[obs_name]
                    noodle = curvature2noodle(self.ref_agent.curvature,
                                              sensor.camera_param,
                                              mode='camera')
                    obs = cv2.polylines(obs, [noodle], False, (255, 0, 0), 2)

                    if not self._config["vis_full_frame"]:
                        # Black out the sides for visualization
                        h, w = obs.shape[:2]
                        h_, w_ = (0.65 * h, 0.65 * w)
                        hs, ws = (int((h - h_) // 2), int((w - w_) // 2))
                        obs = cv2.resize(obs[hs:-hs, ws:-ws], (w, h))

                    obs_render = fit_img_to_ax(self._fig, self._axes[ax_name],
                                               obs[:, :, ::-1])

                elif obs_name in event_cameras.keys():
                    event_cam_param = event_cameras[obs_name].camera_param
                    frame_obs = events2frame(obs, event_cam_param.get_height(),
                                             event_cam_param.get_width())
                    sensor = event_cameras[obs_name]
                    frame_obs = plot_roi(frame_obs.copy(),
                                         sensor.camera_param.get_roi())
                    noodle = curvature2noodle(self.ref_agent.curvature,
                                              sensor.camera_param,
                                              mode='camera')
                    frame_obs = cv2.polylines(frame_obs, [noodle], False,
                                              (0, 0, 255), 2)
                    obs_render = fit_img_to_ax(self._fig, self._axes[ax_name],
                                               frame_obs[:, :, ::-1])

                elif obs_name in lidars.keys():
                    if isinstance(obs, Pointcloud):
                        obs_render = None
                        ax = self._axes[ax_name]
                        ax.clear()
                        obs = obs[::10]  # sub-sample the pointcloud for vis

                        ax, scat = plot_pointcloud(
                            obs,
                            ax=ax,
                            color_by="z",
                            max_dist=20.,
                            car_dims=(self.ref_agent.length,
                                      self.ref_agent.width),
                            cmap="nipy_spectral")

                        # Plot the noodle
                        noodle = curvature2noodle(self.ref_agent.curvature,
                                                  mode='lidar')
                        ax.plot(noodle[:, 0], noodle[:, 1], '-r', linewidth=3)

                    else:  # dense image
                        obs = np.roll(obs, -obs.shape[1] // 4, axis=1)  # shift
                        obs = np.concatenate(np.split(obs, 2, axis=1),
                                             0)  # stack
                        obs = np.clip(4 * obs, 0, 255).astype(np.uint8)  # norm
                        obs = cv2.applyColorMap(obs, cv2.COLORMAP_JET)  # color
                        obs_render = fit_img_to_ax(self._fig,
                                                   self._axes[ax_name], obs)

                else:
                    logging.error(f'Unrecognized observation {obs_name}')
                    continue
                title = '{}: {}'.format(agent.id, obs_name)
                self._axes[ax_name].set_title(title,
                                              color='white',
                                              size=20,
                                              weight='bold')
                if obs_render is not None:
                    self._artists['im:{}'.format(ax_name)].set_data(obs_render)

        # Convert to image
        img = fig2img(self._fig)

        return img

    def _update_patch(self, ax: plt.Axes, name: str,
                      patch: PolygonPatch) -> None:
        if name in self._artists:
            self._artists[name].remove()
        ax.add_patch(patch)
        self._artists[name] = patch

    def _get_timestamp(self, frame_index: int) -> Tuple[float, bool]:
        return self.ref_agent.trace.get_master_timestamp(
            self.ref_agent.segment_index, frame_index, check_end=True)

    @property
    def ref_agent(self) -> Car:
        """ Agent as a reference to compute poses of objects (e.g., cars, road)
            in visualization. """
        return self._world.agents[0]


def curvature2noodle(curvature: float,
                     camera_param: Optional[CameraParams] = None,
                     mode: Optional[str] = 'camera') -> np.ndarray:
    """ Construct a curly line (noodle) based on the curvature for visualizing
    steering control command.

    Args:
        curvature (float): Curvature (steering angle control command).
        camera_param (vista.entities.sensors.camera_utils.CameraParams): Camera parameters; used if
        mode is set to camera.
        mode (str): Sensor type for the visualization.

    Returns:
        np.ndarray: A curly line that visualizes the given curvature.

    Raises:
        NotImplementedError: Unrecognized mode to draw the noodle.

    """
    lookaheads = np.linspace(0, 15, 10)  # meters
    if mode == 'camera':
        assert camera_param is not None

        K = camera_param.get_K()
        normal = camera_param.get_ground_plane()[0:3]
        normal = np.reshape(normal, [1, 3])
        d = camera_param.get_ground_plane()[3]
        A, B, C = normal[0]

        radius = 1. / (curvature + 1e-8)

        z_vals = lookaheads
        y_vals = (d - C * z_vals) / B
        x_sq_r = radius**2 - z_vals**2 - (y_vals - d)**2
        x_vals = np.sqrt(x_sq_r[x_sq_r > 0]) - abs(radius)
        y_vals = y_vals[x_sq_r > 0]
        z_vals = z_vals[x_sq_r > 0]

        if radius < 0:
            x_vals *= -1

        world_coords = np.stack((x_vals, y_vals, z_vals))

        theta = camera_param.get_yaw()
        R = np.array([[np.cos(theta), 0.0, -np.sin(theta)], [0.0, 1.0, 0.0],
                      [np.sin(theta), 0.0, np.cos(theta)]])
        tf_world_coords = np.matmul(R, world_coords)
        img_coords = np.matmul(K, tf_world_coords)
        norm = np.divide(img_coords, img_coords[2] + 1e-10)

        valid_inds = np.multiply(norm[0] >= 0,
                                 norm[0] < camera_param.get_width())
        valid_inds = np.multiply(valid_inds, norm[1] >= 0)
        valid_inds = np.multiply(valid_inds,
                                 norm[1] < camera_param.get_height())

        noodle = norm[:2, valid_inds].astype(np.int32).T
    elif mode == 'lidar':
        turning_r = 1 / (curvature + 1e-8)
        shifts = (np.sqrt(turning_r**2 - lookaheads**2) - abs(turning_r))
        shifts = -1 * np.sign(turning_r) * shifts
        noodle = np.stack([lookaheads, shifts], axis=1)
    else:
        raise NotImplementedError(
            'Unrecognized mode {} in drawing noodle'.format(mode))

    return noodle


def plot_roi(img: np.ndarray,
             roi: List[int],
             color: Optional[List[int]] = (0, 0, 255),
             thickness: Optional[int] = 2) -> np.ndarray:
    """ Plot a bounding box that shows ROI on an image.

    Args:
        img (np.ndarray): An image to be plotted.
        roi (List[int]): Region of interest.
        color (List[int]): Color of the bounding box.
        thickness (int): Thickness of the bounding box.

    Returns:
        np.ndarray: An image with the ROI bounding box.

    """
    (i1, j1, i2, j2) = roi
    img = cv2.rectangle(img, (j1, i1), (j2, i2), color, thickness)
    return img


def events2frame(events: List[np.ndarray],
                 cam_h: int,
                 cam_w: int,
                 positive_color: Optional[List] = [255, 255, 255],
                 negative_color: Optional[List] = [212, 188, 114],
                 mode: Optional[int] = 2) -> np.ndarray:
    """ Convert event data to frame representation.

    Args:
        events (List[np.ndarray]): A list with entries as a collection of positive and
                                   negative events.
        cam_h (int): Height of the frame representation.
        cam_w (int): Width of the frame representation.
        positive_color (List): Color of positive events.
        negative_color (List): Color of negative events.
        mode (int): Mode for colorization.

    Returns:
        np.ndarray: Frame representation of event data.

    """
    if mode == 0:
        frame = np.zeros((cam_h, cam_w, 3), dtype=np.uint8)
        for color, p_events in zip([positive_color, negative_color], events):
            uv = np.concatenate(p_events)[:, :2]
            frame[uv[:, 0], uv[:, 1], :] = color
    elif mode == 1:
        frame_acc = np.zeros((cam_h, cam_w), dtype=np.int8)
        for polarity, p_events in zip([1, -1], events):
            for sub_p_events in p_events:
                uv = sub_p_events[:, :2]
                frame_acc[uv[:, 0], uv[:, 1]] += polarity

        frame = np.zeros((cam_h, cam_w, 3), dtype=np.uint8)
        frame[frame_acc > 0, :] = positive_color
        frame[frame_acc < 0, :] = negative_color
    elif mode == 2:
        frame_abs_acc = np.zeros((cam_h, cam_w), dtype=np.int8)
        frame = np.zeros((cam_h, cam_w, 3), dtype=np.uint8)
        for polarity, p_events in zip([1, -1], events):
            for sub_p_events in p_events:
                uv = sub_p_events[:, :2]
                add_c = np.array(
                    positive_color if polarity > 0 else negative_color)[None,
                                                                        ...]
                cnt = frame_abs_acc[uv[:, 0], uv[:, 1]][:, None]
                frame[uv[:, 0], uv[:, 1]] = (frame[uv[:, 0], uv[:, 1]] * cnt +
                                             add_c) / (cnt + 1)
                frame_abs_acc[uv[:, 0], uv[:, 1]] = cnt[:, 0] + 1
    else:
        raise NotImplementedError('Unknown mode {}'.format(mode))
    return frame


def plot_pointcloud(pcd,
                    color_by="z",
                    max_dist=None,
                    cmap="nipy_spectral",
                    car_dims=None,
                    ax=None,
                    scat=None,
                    s=1):
    """ Convert pointcloud to an image for visualization. """
    if ax is None:
        _, ax = plt.subplots()

    if max_dist is not None:
        pcd = pcd[pcd.dist < (max_dist * np.sqrt(2))]

    if color_by == "z":
        c = pcd.z
        vmin, vmax = (-2.5, 4)
    elif color_by == "intensity":
        c = np.log(1 + pcd.intensity)
        vmin, vmax = (1.7, 4.3)
    else:
        raise ValueError(f"unsupported color {color_by}")

    # Plot points
    if scat is None:
        scat = ax.scatter(pcd.x,
                          pcd.y,
                          c=c,
                          s=s,
                          vmin=vmin,
                          vmax=vmax,
                          cmap=cmap)
    else:
        scat.set_offsets(np.stack([pcd.x, pcd.y], axis=1))
        scat.set_clim(vmin, vmax)
        scat.set_color(getattr(plt.cm, cmap)(scat.norm(c)))

    # Plot car
    if car_dims is not None:
        l_car, w_car = car_dims
        ax.add_patch(
            patches.Rectangle(
                (-l_car / 2, -w_car / 2),
                l_car,
                w_car,
                fill=True  # remove background
            ))

    ax.set_xlim(-max_dist, max_dist)
    ax.set_ylim(-max_dist, max_dist)
    return ax, scat


def fig2img(fig: plt.Figure) -> np.ndarray:
    """ Convert a matplotlib figure to a numpy array. """
    fig.canvas.draw()
    buf = fig.canvas.buffer_rgba()
    img = np.asarray(buf)[:, :, :3]
    return img


def fit_img_to_ax(fig: plt.Figure, ax: plt.Axes,
                  img: np.ndarray) -> np.ndarray:
    """ Fit an image to an axis in a matplotlib figure. """
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    w, h = bbox.width, bbox.height
    img_h, img_w = img.shape[:2]
    new_img_w = img_h * w / h
    new_img_h = img_w * h / w
    d_img_w = new_img_w - img_w
    d_img_h = new_img_h - img_h
    if d_img_h > 0:
        pad_img = np.zeros((int(d_img_h // 2), img_w, 3), dtype=np.uint8)
        new_img = np.concatenate([pad_img, img, pad_img], axis=0)
    elif d_img_w > 0:
        pad_img = np.zeros((img_h, int(d_img_w // 2), 3), dtype=np.uint8)
        new_img = np.concatenate([pad_img, img, pad_img], axis=1)
    else:
        raise ValueError('Something weird happened.')
    return new_img



import torch
from torch import nn
import numpy as np


class LidarModel(torch.nn.Module):
    def __init__(self, layers, filters):
        super().__init__()

        self.unet = UNet(in_channels=3,
                         out_channels=2,
                         layers=layers,
                         filters=filters)

        scale = np.array([50., 255., 1.], dtype=np.float32).reshape(1, 3, 1, 1)

        # Save constants into the state dict of the model (for loading back)
        self.register_buffer('scale', torch.from_numpy(scale))
        self.register_buffer('layers', torch.tensor(layers))
        self.register_buffer('filters', torch.tensor(filters))

    def __call__(self, x):
        scale = self.scale
        x = self.unet(x / scale)
        x = scale[:, :2] * torch.exp(x - torch.log(scale[:, :2]) / 2.)
        return x


class UNet(nn.Module):
    def __init__(self,
                 in_channels,
                 out_channels,
                 layers=2,
                 filters=32,
                 kernel=3,
                 pad=1):
        super().__init__()

        # Convolutional
        self.conv_down = [
            self.contract_block(in_channels, filters, kernel, pad)
        ]
        filt = filters
        for i in range(layers):
            block = self.contract_block(filt, 2 * filt, kernel, pad)
            self.conv_down.append(block)
            filt *= 2
        self.conv_down = nn.ModuleList(self.conv_down)

        # Deconvolutional
        self.conv_up = [self.expand_block(filt, filt // 2, kernel, pad)]
        filt //= 2
        for i in range(layers - 1):
            block = self.expand_block(filt * 2, filt // 2, kernel, pad)
            self.conv_up.append(block)
            filt //= 2
        head = self.expand_block(filt * 2, out_channels, kernel, pad)
        self.conv_up.append(head)
        self.conv_up = nn.ModuleList(self.conv_up)

        # Final layer with no activation
        self.final = torch.nn.Conv2d(out_channels, out_channels, 3, padding=1)

    def __call__(self, x):
        # Downsampling
        outs = []
        for layer in self.conv_down:
            x = layer(x)
            outs.append(x)
        # conv1 = self.conv1(x)
        # conv2 = self.conv2(conv1)
        # conv3 = self.conv3(conv2)

        # Upsampling
        x = self.conv_up[0](x)
        # upconv3 = self.upconv3(conv3)

        for layer, skip in zip(self.conv_up[1:], outs[:-1][::-1]):
            clip = torch.tensor(x.shape)[-2:] - torch.tensor(skip.shape)[-2:]
            x = x[:, :, clip[0]:, clip[1]:].clone() # make shape of x match skip
            x = layer(torch.cat([x, skip], 1))
        # upconv2 = self.upconv2(torch.cat([upconv3, conv2], 1))
        # upconv1 = self.upconv1(torch.cat([upconv2, conv1], 1))
        x = self.final(x)

        return x
        # return upconv1

    def contract_block(self, in_channels, out_channels, kernel_size, padding):

        contract = nn.Sequential(
            torch.nn.Conv2d(in_channels,
                            out_channels,
                            kernel_size=kernel_size,
                            stride=1,
                            padding=padding),
            torch.nn.BatchNorm2d(out_channels), torch.nn.LeakyReLU(),
            torch.nn.Conv2d(out_channels,
                            out_channels,
                            kernel_size=kernel_size,
                            stride=1,
                            padding=padding),
            torch.nn.BatchNorm2d(out_channels), torch.nn.LeakyReLU(),
            torch.nn.MaxPool2d(kernel_size=3, stride=2, padding=1))

        return contract

    def expand_block(self, in_channels, out_channels, kernel_size, padding):

        expand = nn.Sequential(
            torch.nn.Conv2d(in_channels,
                            out_channels,
                            kernel_size,
                            stride=1,
                            padding=padding),
            torch.nn.BatchNorm2d(out_channels), torch.nn.LeakyReLU(),
            torch.nn.Conv2d(out_channels,
                            out_channels,
                            kernel_size,
                            stride=1,
                            padding=padding),
            torch.nn.BatchNorm2d(out_channels), torch.nn.LeakyReLU(),
            torch.nn.ConvTranspose2d(out_channels,
                                     out_channels,
                                     kernel_size=3,
                                     stride=2,
                                     padding=1,
                                     output_padding=1))
        return expand



import os
import numpy as np
from typing import Dict, List, Optional
from collections import OrderedDict
import h5py

from . import TopicNames
from ...utils import logging


class MultiSensor:
    """ This class handles synchronization across multiple sensors in a trace. It basically
    reads timestamps files associated with every sensors existing in the trace and construct
    several helper functions for conversion between an unified timestamp used in the simulator
    and timestamp (or frame number, etc) of different sensors.

    Args:
        trace_dir (str): Directory to a trace.
        master_sensor (str): Name of the master sensor.

    Raises:
        AssertionError: (1) The name of the master sensor not included in a predefined
        set of topic names specified in  (2) No timestamp data for the master sensor.

    """
    def __init__(
            self,
            trace_dir: str,
            master_sensor: Optional[str] = TopicNames.master_topic) -> None:
        self._trace_dir: str = trace_dir
        self._master_sensor: str = master_sensor

        # Get frame-to-timestamp mapping for every sensors
        self._sensor_frame_to_time: Dict = dict()
        sensor_topic_names = [TopicNames.lidar_3d] + list(TopicNames.cameras)
        assert master_sensor in sensor_topic_names, \
            f'Master sensor {master_sensor} not in topic names' + \
            f'{sensor_topic_names}. Please check camera config or TopicNames.py'
        for fname in os.listdir(self._trace_dir):
            sensor_name, ext = os.path.splitext(fname)
            if sensor_name in sensor_topic_names:
                fpath = os.path.join(self._trace_dir, fname)
                if ext == '.csv':
                    data = np.genfromtxt(fpath,
                                         delimiter=',',
                                         skip_header=1,
                                         dtype=np.float64)
                    frames, times = (data[:, 0], data[:, 1])
                elif ext == ".h5":
                    f = h5py.File(fpath, "r")
                    times = f["timestamp"][:, 0]
                    frames = np.arange(times.shape[0])
                else:
                    continue  # Not implemented yet...

                frame_to_time = OrderedDict()
                for i in range(len(frames)):
                    frame_to_time[int(frames[i])] = times[i]
                self._sensor_frame_to_time[sensor_name] = frame_to_time

        self._sensor_names: List[str] = list(self._sensor_frame_to_time.keys())
        assert master_sensor in self.sensor_names, \
            f'No timestamp data for the master sensor {master_sensor}'

    def get_time_from_frame_num(self, sensor: str, frame_num: int) -> float:
        """ Compute the timestamp associated to a frame in a video return
        None if we dont have information about that frame.

        Args:
            sensor (str): Sensor name.
            frame_num (int): Frame number.

        Returns:
            float: Timestamp associated with the given sensor and frame number.
        """
        return self._sensor_frame_to_time[sensor].get(frame_num, None)

    def get_frames_from_times(
            self,
            timestamps: List[float],
            fetch_smaller: Optional[bool] = False) -> Dict[str, List[int]]:
        """ Takes in a list of timestamps and returns corresponding frame
        numbers for each sensor. Note that since sensors are not necessarily
        sync'ed, the returned frame numbers are the one with the closest
        (smaller) timestamps.

        Args:
            timestamps (list): A list of timestamps.
            fetch_smaller (bool): Whether to fetch the closes and smaller timestamps.

        Returns:
            dict: Corresponding frame numbers for all sensors.
        """
        frames = dict()
        timestamps = np.array(timestamps)
        for sensor in self.sensor_names:
            frame_to_time = self._sensor_frame_to_time[sensor]

            frames[sensor] = []
            pointer = 0
            for ts in timestamps:
                while pointer < len(frame_to_time) - 1:
                    if ts >= frame_to_time[pointer] and ts < frame_to_time[
                            pointer + 1]:
                        if fetch_smaller:
                            frames[sensor].append(pointer)
                        else:
                            if np.abs(frame_to_time[pointer] - ts) >= \
                                np.abs(ts - frame_to_time[pointer + 1]):
                                frames[sensor].append(pointer + 1)
                            else:
                                frames[sensor].append(pointer)
                        break
                    else:
                        pointer += 1

        return frames

    def get_master_timestamps(self) -> List[float]:
        """ Get all timestamps of the main sensor.

        Returns:
            list: A list of timestamp.
        """
        # using values() works since it's an ordered dict
        timestamps = list(
            self._sensor_frame_to_time[self._master_sensor].values())
        return timestamps

    def set_main_sensor(self, sensor_type: str, sensor_name: str) -> None:
        """ Set main sensor based on sensor's type and name.

        Args:
            sensor_type (str): Type of the sensor to be set (camera, lidar, or event camera).
            sensor_name (str): Name of the sensor to be set.
        """
        assert sensor_type in ['camera', 'lidar', 'event_camera']
        setattr(self, '_main_{}'.format(sensor_type), sensor_name)

    @property
    def sensor_names(self) -> List[str]:
        """ A list of all sensors' names. """
        return self._sensor_names

    @property
    def camera_names(self) -> List[str]:
        """ A list of RGB cameras' names. """
        logging.debug('Hacky way to include RGB camera with name front_center')
        return [
            _x for _x in self._sensor_names
            if 'camera' in _x or _x == 'front_center'
        ]

    @property
    def main_camera(self) -> str:
        """ The main RGB camera object. """
        return self._main_camera if hasattr(self, '_main_camera') else None

    @property
    def lidar_names(self) -> List[str]:
        """ A list of LiDARs' names. """
        return [_x for _x in self._sensor_names if 'lidar' in _x]

    @property
    def main_lidar(self) -> str:
        """ The main LiDAR object """
        return self._main_lidar if hasattr(self, '_main_lidar') else None

    @property
    def main_event_camera(self) -> str:
        """ The main event camera object. """
        return self._main_event_camera if hasattr(
            self, '_main_event_camera') else None

    @property
    def master_sensor(self) -> str:
        """ The name of the master sensor. """
        return self._master_sensor



from typing import List, Dict, Any, Optional, Callable
from vista.entities.sensors.EventCamera import EventCamera
import numpy as np
from collections import deque

from .Dynamics import State, StateDynamics, curvature2steering, curvature2tireangle, \
                      tireangle2curvature, update_with_perfect_controller
from ..Entity import Entity

from ..sensors import BaseSensor, Camera, Lidar, EventCamera
from ...core import World, Trace
from ...utils import transform, logging, misc


class Car(Entity):
    """ The class of a car agent. This object lives in the :class:`World` and is attached
    to a trace object that provides pointers to dataset for data-driven simulation and zero
    or one or more sensor objects that synthesize sensory measurement. The update of vehicle
    state is handled in this object.

    Args:
        world (World): The world that this agent lives in.
        car_config (Dict): Configuration of the car. An example (default) is,

            >>> DEFAULT_CONFIG = {
                'length': 5.,
                'width': 2.,
                'wheel_base': 2.78,
                'steering_ratio': 14.7,
                'lookahead_road': False,
                'road_buffer_size': 200,
            }

    Example Usage (always make sure reset is called first to initialize vehicle state
    or pointer to the dataset for data-driven simulation)::

        >>> car = world.spawn_agent(car_config)
        >>> car.spawn_camera(camera_config)
        >>> world.reset()
        >>> car.step_dynamics(action) # update vehicle states
        >>> car.step_sensors() # do sensor capture
        >>> observation = car.observations # fetch sensory measurement
        >>> car.step_dataset() # simply get next frame in the dataset without synthesis

    """
    DEFAULT_CONFIG = {
        'length': 5.,
        'width': 2.,
        'wheel_base': 2.78,
        'steering_ratio': 14.7,
        'lookahead_road': False,
        'road_buffer_size': 200,
    }

    def __init__(self, world: World, car_config: Dict) -> None:
        super(Car, self).__init__()

        # Car configuration
        car_config = misc.merge_dict(car_config, self.DEFAULT_CONFIG)
        self._config = car_config

        # Pointer to the parent vista.World object where the agent lives
        self._parent: World = world

        # The trace to be associated with, updated every reset
        self._trace: Trace = None

        # A list of sensors attached to this agent (List[vista.Sensor]).
        self._sensors: List[BaseSensor] = []

        # State dynamics for tracking the virtual and human agent
        self._relative_state: State = State()
        self._ego_dynamics: StateDynamics = StateDynamics()
        self._human_dynamics: StateDynamics = StateDynamics()

        # Properties of a car
        self._length: float = self.config['length']
        self._width: float = self.config['width']
        self._wheel_base: float = self.config['wheel_base']
        self._steering_ratio: float = self.config['steering_ratio']
        self._speed: float = 0.
        self._curvature: float = 0.
        self._steering: float = 0.
        self._tire_angle: float = 0.
        self._human_speed: float = 0.
        self._human_curvature: float = 0.
        self._human_steering: float = 0.
        self._human_tire_angle: float = 0.
        self._timestamp: float = 0.
        self._frame_number: int = 0
        self._trace_index: int = 0
        self._segment_index: int = 0
        self._frame_index: int = 0
        self._observations: Dict[str, Any] = dict()
        self._done: bool = False

        # Privileged information
        if self.config['lookahead_road']:
            road_buffer_size = self.config['road_buffer_size']
            self._road: deque[np.ndarray] = deque(maxlen=road_buffer_size)
            self._road_frame_idcs: deque[int] = deque(maxlen=road_buffer_size)
            self._road_dynamics: StateDynamics = StateDynamics()

    def spawn_camera(self, config: Dict) -> Camera:
        """ Spawn and attach a camera to this car.

        Args:
            config (Dict): Configuration of camera and rendering. For
                           more details, please check the doc of
                           :class:`Camera` sensor.

        Returns:
            Camera: a vista camera sensor object spawned.

        """
        name = config.get('name', None)
        logging.info(f'Spawn a new camera {name} in car ({self.id})')
        cam = Camera(attach_to=self, config=config)
        self._sensors.append(cam)

        return cam

    def spawn_lidar(self, config: Dict) -> Lidar:
        """ Spawn and attach a LiDAR to this car.

        Args:
            config (Dict): Configuration of LiDAR. For more details,
                           please check the doc of :class:`Lidar` sensor.

        Returns:
            Lidar: a vista Lidar sensor object spawned.

        """
        name = config.get('name', None)
        logging.info(f'Spawn a new lidar {name} in car ({self.id})')
        lidar = Lidar(attach_to=self, config=config)
        self._sensors.append(lidar)

        return lidar

    def spawn_event_camera(self, config: Dict) -> EventCamera:
        """ Spawn and attach an event camera to this car.

        Args:
            config (Dict): Configuration of event camera. For more details,
                           please check the doc of :class:`EventCamera` sensor.

        Returns:
            EventCamera: a vista event camera sensor object spawned.

        """
        name = config.get('name', None)
        logging.info(f'Spawn a new event camera {name} in car ({self.id})')
        event_cam = EventCamera(attach_to=self, config=config)
        self._sensors.append(event_cam)

        return event_cam

    def reset(self,
              trace_index: int,
              segment_index: int,
              frame_index: int,
              initial_dynamics_fn: Optional[Callable] = None,
              step_sensors: Optional[bool] = True) -> None:
        """ Reset the car. This involves pointing to somewhere in the dataset for later-on
        data-driven simulation, initializing vehicle state, and resetting all sensors attached
        to this car. If ``lookahead_road = True``, the road cache will also be reset.

        Args:
            trace_index (int): A pointer to which trace to be simulated on.
            segment_index (int): A pointer to which segment in a trace to be simulated on.
            frame_index (int): A pointer to which frame in a segment to be simulated on.
            initial_dynamics_fn (Callable): A function to initialize vehicle state. The
                function takes x, y, yaw, steering (tire angle), and speed as inputs. Default
                is set to ``None``, which initialize vehicle with the same state as the dataset.
            step_sensors (bool): Whether to step sensor; default is set to ``True``.

        """
        logging.info(f'Car ({self.id}) reset')

        # Update pointers to dataset
        self._trace = self.parent.traces[trace_index]
        self._timestamp = self.trace.get_master_timestamp(
            segment_index, frame_index)
        self._frame_number = self.trace.get_master_frame_number(
            segment_index, frame_index)
        self._trace_index = trace_index
        self._segment_index = segment_index
        self._frame_index = frame_index

        self._done = False

        # Reset states and dynamics
        self._human_speed = self.trace.f_speed(self.timestamp)
        self._human_curvature = self.trace.f_curvature(self.timestamp)
        self._human_steering = curvature2steering(self.human_curvature,
                                                  self.wheel_base,
                                                  self.steering_ratio)
        self._human_tire_angle = curvature2tireangle(self.human_curvature,
                                                     self.wheel_base)
        self.human_dynamics.update(0., 0., 0., self.human_tire_angle,
                                   self.human_speed)

        initial_dynamics = self.human_dynamics.numpy().copy()
        if initial_dynamics_fn is not None:
            initial_dynamics = initial_dynamics_fn(*initial_dynamics)
        self.ego_dynamics.update(*initial_dynamics)
        self._speed = self.human_speed
        self._curvature = self.human_curvature
        self._steering = self.human_steering
        self._tire_angle = curvature2tireangle(self.curvature, self.wheel_base)

        latlongyaw = transform.compute_relative_latlongyaw(
            self.ego_dynamics.numpy()[:3],
            self.human_dynamics.numpy()[:3])
        self._relative_state.update(*latlongyaw)

        # Reset for privileged information
        if hasattr(self, '_road'):
            self._road.clear()
            self._road.append(self.human_dynamics.numpy())
            self._road_dynamics = self.human_dynamics.copy()
            self._road_frame_idcs.clear()
            self._road_frame_idcs.append(self.frame_index)
            self._update_road()

        # Reset sensors
        for sensor in self.sensors:
            if isinstance(
                    sensor,
                    Camera) and self._trace.multi_sensor.main_camera is None:
                self._trace.multi_sensor.set_main_sensor('camera', sensor.name)
            elif isinstance(
                    sensor,
                    Lidar) and self._trace.multi_sensor.main_lidar is None:
                self._trace.multi_sensor.set_main_sensor('lidar', sensor.name)
            elif isinstance(
                    sensor, EventCamera
            ) and self._trace.multi_sensor.main_event_camera is None:
                self._trace.multi_sensor.set_main_sensor(
                    'event_camera', sensor.name)

            sensor.reset()

        # Update observation
        if step_sensors:
            self.step_sensors()

    def step_dataset(self, step_dynamics=True):
        """ Step through the dataset without rendering. This is basically
        fetching the next frame from the dataset. Normally, it is called
        when doing imitation learning.

        Args:
            step_dynamics (bool): Whether to update vehicle state; default
                is set to ``True``.

        Raises:
            NotImplementedError: if any attached sensor has no implemented
                function for stepping through dataset.

        """
        logging.info(f'Car ({self.id}) step based on dataset')

        # Step by incrementing frame number
        ts = self.timestamp
        frame_index = self.frame_index + 1
        exceed_end, self._timestamp = self.trace.get_master_timestamp(
            self.segment_index, frame_index, check_end=True)
        if exceed_end:  # trigger trace done terminatal condition
            self._done = True
            logging.info(f'Car ({self.id}) exceed the end of trace')
        else:
            self._frame_index = frame_index
            self._frame_number = self.trace.get_master_frame_number(
                self.segment_index, self.frame_index)

            self._human_speed = self.trace.f_speed(self.timestamp)
            self._human_curvature = self.trace.f_curvature(self.timestamp)
            self._human_steering = curvature2steering(self.human_curvature,
                                                      self.wheel_base,
                                                      self.steering_ratio)
            self._human_tire_angle = curvature2tireangle(
                self.human_curvature, self.wheel_base)

            self._speed = self._human_speed
            self._curvature = self._human_curvature
            self._steering = self._human_steering
            self._tire_angle = self._human_tire_angle

            # Step human dynamics if queried
            if step_dynamics:
                current_state = [
                    curvature2tireangle(self.human_curvature, self.wheel_base),
                    self.human_speed
                ]
                update_with_perfect_controller(current_state,
                                               self.timestamp - ts,
                                               self._human_dynamics)
                self._ego_dynamics = self.human_dynamics.copy()

            # Get image frame
            self._observations = dict()
            for sensor in self.sensors:
                if type(sensor) not in [Camera, Lidar, EventCamera]:
                    raise NotImplementedError(
                        f'Sensor {sensor} is not supported in step dataset')
                self._observations[sensor.name] = sensor.capture(
                    self.timestamp)

    def step_dynamics(self,
                      action: np.ndarray,
                      dt: Optional[float] = 1 / 30.,
                      update_road: Optional[bool] = True) -> None:
        """ Update vehicle state given control command based on vehicle dynamics
        and update timestamp, which is then used to update pointer to the dataset
        for data-driven simulation.

        Args:
            action (np.ndarray): Control command (curvature and speed).
            dt (float): Elapsed time.
            update_road (bool): Whether to update road cache; default is
                set to ``True``.

        """
        assert not self.done, 'Agent status is done. Please call reset first.'
        logging.info(f'Car ({self.id}) step dynamics')

        # Parse action
        action = np.array(action).reshape(-1)
        assert action.shape[0] == 2
        desired_curvature, desired_speed = action
        desired_tire_angle = curvature2tireangle(desired_curvature,
                                                 self.wheel_base)

        # Run low-level controller and step vehicle dynamics
        # TODO: non-perfect low-level controller
        logging.debug('Using perfect low-level controller now')
        desired_state = [desired_tire_angle, desired_speed]
        update_with_perfect_controller(desired_state, dt, self._ego_dynamics)

        # Update based on vehicle dynamics feedback
        self._tire_angle = self.ego_dynamics.steering
        self._speed = self.ego_dynamics.speed
        self._curvature = tireangle2curvature(self.tire_angle, self.wheel_base)
        self._steering = curvature2steering(self.curvature, self.wheel_base,
                                            self.steering_ratio)

        # Update human (reference) dynamics for assoication with the trace / dataset
        human = self.human_dynamics.copy()
        top2_closest = dict(dist=deque([np.inf, np.inf], maxlen=2),
                            dynamics=deque([None, None], maxlen=2),
                            timestamp=deque([None, None], maxlen=2),
                            index=deque([None, None], maxlen=2))
        index = self.frame_index
        ts = self.trace.get_master_timestamp(self.segment_index, index)
        while True:
            dist = np.linalg.norm(human.numpy()[:2] -
                                  self.ego_dynamics.numpy()[:2])
            if dist < top2_closest['dist'][1]:
                if dist < top2_closest['dist'][0]:
                    top2_closest['dist'].appendleft(dist)
                    top2_closest['dynamics'].appendleft(human.copy())
                    top2_closest['timestamp'].appendleft(ts)
                    top2_closest['index'].appendleft(index)
                else:
                    top2_closest['dist'][1] = dist
                    top2_closest['dynamics'][1] = human.copy()
                    top2_closest['timestamp'][1] = ts
                    top2_closest['index'][1] = index
            else:
                break

            next_index = index + 1 * int(np.sign(dt))
            exceed_end, next_ts = self.trace.get_master_timestamp(
                self.segment_index, next_index, check_end=True)
            if exceed_end:  # trigger trace done terminatal condition
                self._done = True
                logging.info(f'Car ({self.id}) exceed the end of trace')

            current_state = [
                curvature2tireangle(self.trace.f_curvature(ts),
                                    self.wheel_base),
                self.trace.f_speed(ts)
            ]
            update_with_perfect_controller(current_state, next_ts - ts, human)

            index = next_index
            ts = next_ts

        self._human_dynamics = top2_closest['dynamics'][0].copy()
        self._frame_index = top2_closest['index'][0]
        self._frame_number = self.trace.get_master_frame_number(
            self.segment_index, self.frame_index)

        # Update timestamp based on where the car position is with respect to
        # course distance of trace (may not be exactly the same as any of
        # timestamps in the dataset)
        latlongyaw_closest = transform.compute_relative_latlongyaw(
            self.ego_dynamics.numpy()[:3],
            top2_closest['dynamics'][0].numpy()[:3])
        latlongyaw_second_closest = transform.compute_relative_latlongyaw(
            self.ego_dynamics.numpy()[:3],
            top2_closest['dynamics'][1].numpy()[:3])
        ratio = abs(latlongyaw_second_closest[1]) / (
            abs(latlongyaw_closest[1]) + abs(latlongyaw_second_closest[1]))
        self._timestamp = (ratio * top2_closest['timestamp'][0] +
                           (1. - ratio) * top2_closest['timestamp'][1])

        # Update human control based on current timestamp
        self._human_speed = self.trace.f_speed(self.timestamp)
        self._human_curvature = self.trace.f_curvature(self.timestamp)
        self._human_steering = curvature2steering(self.human_curvature,
                                                  self.wheel_base,
                                                  self.steering_ratio)
        self._human_tire_angle = curvature2tireangle(self.human_curvature,
                                                     self.wheel_base)

        # Update relative transformation between human and ego dynamics
        self._relative_state.update(*latlongyaw_closest)

        # Update privileged information (in global coordinates)
        if update_road and hasattr(self, '_road'):
            self._update_road()

    def step_sensors(self) -> None:
        """ Update sensor measurement given current state of the vehicle. """
        logging.info(f'Car ({self.id}) step sensors')
        self._observations = dict()
        for sensor in self.sensors:
            self._observations[sensor.name] = sensor.capture(self.timestamp)

    def _update_road(self) -> None:
        exceed_end = False
        get_timestamp = lambda _idx: self.trace.get_master_timestamp(
            self.segment_index, _idx, check_end=True)
        while self._road_frame_idcs[-1] < (
                self.frame_index + self._road.maxlen / 2.) and not exceed_end:
            exceed_end, ts = get_timestamp(self._road_frame_idcs[-1])
            self._road_frame_idcs.append(self._road_frame_idcs[-1] + 1)
            exceed_end, next_ts = get_timestamp(self._road_frame_idcs[-1])

            state = [
                curvature2tireangle(self.trace.f_curvature(ts),
                                    self.wheel_base),
                self.trace.f_speed(ts)
            ]
            update_with_perfect_controller(state, next_ts - ts,
                                           self._road_dynamics)
            self._road.append(self._road_dynamics.numpy())

    @property
    def trace(self) -> Trace:
        """ The :class:`Trace` currently associated with the car. """
        return self._trace

    @property
    def sensors(self) -> List[BaseSensor]:
        """ All sensors attached to this car. """
        return self._sensors

    @property
    def relative_state(self) -> State:
        """ Relative transform between ``ego_dynamics`` and ``human_dynamics``. """
        return self._relative_state

    @property
    def ego_dynamics(self) -> StateDynamics:
        """ Current simulated vehicle state. """
        return self._ego_dynamics

    @property
    def human_dynamics(self) -> StateDynamics:
        """ Vehicle state of the current pointer to the dataset (human trajectory). """
        return self._human_dynamics

    @property
    def length(self) -> float:
        """ Car length. """
        return self._length

    @property
    def width(self) -> float:
        """ Car width. """
        return self._width

    @property
    def wheel_base(self) -> float:
        """ Wheel base. """
        return self._wheel_base

    @property
    def steering_ratio(self) -> float:
        """ Steering ratio. """
        return self._steering_ratio

    @property
    def speed(self) -> float:
        """ Speed of simulated trajectory (this car) in current timestamp. """
        return self._speed

    @property
    def curvature(self) -> float:
        """ Curvature of simulated trajectory (this car) in current timestamp. """
        return self._curvature

    @property
    def steering(self) -> float:
        """ Steering angle of simulated trajectory (this car) in current timestamp. """
        return self._steering

    @property
    def tire_angle(self) -> float:
        """ Tire angle of simulated trajectory (this car) in current timestamp. """
        return self._tire_angle

    @property
    def human_speed(self) -> float:
        """ Speed of human trajectory in current timestamp. """
        return self._human_speed

    @property
    def human_curvature(self) -> float:
        """ Curvature of human trajectory in current timestamp. """
        return self._human_curvature

    @property
    def human_steering(self) -> float:
        """ Steering angle of human trajectory in current timestamp. """
        return self._human_steering

    @property
    def human_tire_angle(self) -> float:
        """ Tire angle of human trajectory in current timestamp. """
        return self._human_tire_angle

    @property
    def timestamp(self) -> float:
        """ Current timestamp (normally ROS timestamp). This serves as a
        continuous pointer to the dataset as opposed to ``trace_index``,
        ``segment_index``, and ``frame_index``. """
        return self._timestamp

    @property
    def frame_number(self) -> int:
        """ Current frame number. Note that this is different from ``frame_index``
        as it is a different pointer based on how we define frame in the ``master_sensor``
        instead of a pointer to the dataset. There is only one unique pointer to the dataset,
        which can be mapped to (potentially) different pointers to the frame number in
        different sensors. """
        return self._frame_number

    @property
    def trace_index(self) -> int:
        """ Current pointer to the trace. """
        return self._trace_index

    @property
    def segment_index(self) -> int:
        """ Current pointer to the segment in the current trace. """
        return self._segment_index

    @property
    def frame_index(self) -> int:
        """ Current pointer to the frame in the current segment. """
        return self._frame_index

    @property
    def observations(self) -> Dict[str, Any]:
        """ Sensory measurement at current timestamp. """
        return self._observations

    @property
    def done(self) -> bool:
        """ Whether exceeding the end of the trace currently associated with the car. """
        return self._done

    @property
    def road(self) -> np.ndarray:
        """ Road cache if ``lookahead_road = True`` otherwise ``None``. """
        return np.array(self._road) if hasattr(self, '_road') else None

    @property
    def config(self) -> Dict:
        """ Configuration of this car. """
        return self._config

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} (id={self.id})> ' + \
               f'width: {self.width} ' + \
               f'length: {self.length} ' + \
               f'wheel_base: {self.wheel_base} ' + \
               f'steering_ratio: {self.steering_ratio} ' + \
               f'speed: {self.speed} ' + \
               f'curvature: {self.curvature} '



import os
from typing import Dict, Optional, Tuple, List
from enum import Enum
import numpy as np
import pyrender
import copy

from . import CameraParams
from ....utils import transform, logging, misc

ZNEAR = 0.01
ZFAR = 10000


class DepthModes(Enum):
    FIXED_PLANE = 1
    INPUT_DISP = 2
    MONODEPTH = 3


class ViewSynthesis:
    """ A RGB synthesizer that simulates RGB image at novel viewpoint around a pre-collected
    RGB image/video dataset. Conceptually, it (1) projects a reference 2D RGB image to 3D colored
    mesh using camera projection matrices and (approximated) depth, (2) place virtual objects
    in the scene, (3) render RGB image at novel viewpoint.

    Args:
        camera_param (CameraParams): Camera parameter object of the virtual camera.
        config (Dict): Configuration of the synthesizer.
        init_with_bg_mesh (bool): Whether to initialize with background mesh; default
            is set to ``True``.

    """
    DEFAULT_CONFIG = {
        'depth_mode': DepthModes.FIXED_PLANE,
        'znear': ZNEAR,
        'zfar': ZFAR,
        'use_lighting': False,
        'recoloring_factor': 0.5,
    }

    def __init__(self,
                 camera_param: CameraParams,
                 config: Dict,
                 init_with_bg_mesh: Optional[bool] = True) -> None:
        # Parse configuration
        self._camera_param = camera_param
        config = misc.merge_dict(config, self.DEFAULT_CONFIG)
        self._config = config
        if isinstance(self._config['depth_mode'], str):
            self._config['depth_mode'] = getattr(DepthModes,
                                                 self._config['depth_mode'])

        # Renderer and scene
        self._renderer = pyrender.OffscreenRenderer(
            self._camera_param.get_width(), self._camera_param.get_height())
        self._scene = pyrender.Scene(ambient_light=[1., 1., 1.],
                                     bg_color=[0, 0, 0])

        # Camera for rendering
        cam_w = camera_param.get_width()
        cam_h = camera_param.get_height()
        camera = pyrender.IntrinsicsCamera(
            fx=self._camera_param._fx,
            fy=self._camera_param._fy,
            cx=cam_w / 2.,  # NOTE: assume calibrated camera
            cy=cam_h / 2.,
            znear=self._config['znear'],
            zfar=self._config['zfar'])
        self._camera_node = pyrender.Node(
            name='camera',
            camera=camera,
            translation=self._camera_param.get_position()[:, 0],
            rotation=self._camera_param.get_quaternion()[:, 0])
        self._scene.add_node(self._camera_node)

        # Mesh of background. Can add more by calling add_bg_mesh for different camera_param
        self._world_rays: Dict[str, np.ndarray] = dict()
        if self._config['depth_mode'] == DepthModes.FIXED_PLANE:
            self._depth: Dict[str, np.ndarray] = dict()
        self._bg_node: Dict[str, pyrender.Node] = dict()
        if init_with_bg_mesh:
            self.add_bg_mesh(self._camera_param)

        # Scene for non-background objects and object nodes
        self._object_scene = pyrender.Scene(ambient_light=[1., 1., 1.],
                                            bg_color=[0, 0, 0])
        self._object_nodes = dict()

    def synthesize(
        self,
        trans: np.ndarray,
        rot: np.ndarray,
        imgs: Dict[str, np.ndarray],
        depth: Optional[Dict[str, np.ndarray]] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """ Synthesize RGB image at the novel viewpoint specified by ``trans`` and
        ``rot`` with respect to the nominal viewpoint that corresponds to a set of
        RGB images and depth maps.

        Args:
            trans (np.ndarray): Translation vector.
            rot (np.ndarray): Rotation vector in Euler angle.
            imgs (Dict[str, np.ndarray]): A set of images (potentially from multiple camera).
            depth (Dict[str, np.ndarray]): A set of depth maps corresponding to ``imgs``.

        Returns:
            Returns a tuple (``array_1``, ``array_2``), where ``array_1`` is the
            synthesized RGB image and ``array_2`` is the corresponding depth image.

        """

        for name, img in imgs.items():
            # Need this otherwise will cause memory leak
            if os.environ.get('PYOPENGL_PLATFORM', None) != 'egl':
                self._bg_node[name].mesh = copy.deepcopy(
                    self._bg_node[name].mesh)

            # Refresh meshes in renderer; otherwise mesh vertex/color won't update
            mesh = self._bg_node[name].mesh
            for prim in mesh.primitives:
                prim._unbind()
                prim._remove_from_context()

            if mesh in self._renderer._renderer._meshes:
                self._renderer._renderer._meshes.remove(mesh)

            # Update mesh vertex positions (this won't affect node-level translation and rotation)
            if self._config['depth_mode'] == DepthModes.FIXED_PLANE:
                depth = self._depth[name]
            else:
                raise NotImplementedError

            depth = depth.reshape((1, -1))
            world_coords = np.multiply(-depth, self._world_rays[name])
            mesh.primitives[0].positions = world_coords.T

            # Update mesh face colors
            colors = img[:, ::-1] / 255.
            mesh.primitives[0].color_0[:, :3] = colors.reshape((-1, 3))

        # Update camera pose based on the requested viewpoint (with camera
        # relative pose to the car as offset)
        trans += self._camera_param.get_position()[:, 0]
        rot += transform.quat2euler(self._camera_param.get_quaternion()[:, 0])
        self._camera_node.matrix = transform.vec2mat(trans, rot)

        # Render background
        color_bg, _ = self._renderer.render(
            self._scene, flags=pyrender.constants.RenderFlags.FLAT)

        # Render scene object
        self._object_scene.clear()

        light = pyrender.DirectionalLight(
            [255, 255, 255], self.config['directional_light_intensity'])
        self._object_scene.add(light)

        self._object_scene.add_node(self._camera_node)

        for name, object_node in self._object_nodes.items():
            self._object_scene.add_node(object_node)

        color_objects, depth_objects = self._renderer.render(
            self._object_scene)
        color_objects = color_objects[:, :, ::
                                      -1]  # TODO: not sure why we need this

        # Overlay
        mask = (depth_objects != 0.).astype(np.uint8)[:, :, None]

        if mask.sum() != 0:  # recoloring
            color_bg_mean = color_bg.mean(0).mean(0)
            color_objects_mean = (color_objects *
                                  mask).sum(0).sum(0) / mask.sum()
            color_objects = color_objects + (
                color_bg_mean -
                color_objects_mean) * self.config['recoloring_factor']
            color_objects = np.clip(color_objects, 0, 255).astype(np.uint8)
        else:  # agent out-of-view
            pass

        color = (1 - mask) * color_bg + mask * color_objects

        return color, depth

    def update_object_node(self, name: str, mesh: pyrender.Mesh,
                           trans: np.ndarray, quat: np.ndarray) -> None:
        """ Update the virtual object in the scene.

        Args:
            name (str): Name of the virtual object.
            mesh (pyrender.Mesh): Mesh of the virtual object.
            trans (np.ndarray): Translation vector.
            quat (np.ndarray): Quaternion vector.

        """
        self._object_nodes[name] = pyrender.Node(name=name,
                                                 mesh=mesh,
                                                 translation=trans,
                                                 rotation=quat)

    def add_bg_mesh(self, camera_param: CameraParams) -> None:
        """ Add background mesh to the scene based on camera projection and the
        initial depth. The color of the mesh will be updated at every ``synthesize``
        call and if not using ground-plane depth approximation, the geometry of
        the mesh will be also updated.

        Args:
            camera_param (CameraParams): Camera parameter of the virtual camera.

        """
        # Projection and re-projection parameters
        K = camera_param.get_K().copy()
        logging.debug('Hacky way to handle unrectified image')
        cam_w = camera_param.get_width()
        cam_h = camera_param.get_height()
        K[0, 2] = cam_w / 2.
        K[1, 2] = cam_h / 2.
        K_inv = np.linalg.inv(K)

        # Mesh coordinates, faces, and rays
        name = camera_param.name
        homo_coords, mesh_faces = self._get_homogeneous_image_coords(
            camera_param, get_mesh=True)
        self._world_rays[name] = np.matmul(K_inv, homo_coords)

        # Get depth for ground plane assumption
        if self._config['depth_mode'] == DepthModes.FIXED_PLANE:
            normal = np.reshape(camera_param.get_ground_plane()[:3], [1, 3])
            d = camera_param.get_ground_plane()[3]
            k = np.divide(d, np.matmul(normal, self._world_rays[name]))
            k[k < 0] = self._config[
                'zfar'] / 10.  # should be smaller than actual zfar
            if camera_param == self._camera_param:
                # NOTE: hacky way to make image from the main camera have fronter order
                logging.debug(
                    'Hacky way to make main camera image have fronter order')
                k = k / 1.1
            self._depth[camera_param.name] = k

        # Add mesh to scene (fix node level translation and rotation)
        mesh = pyrender.Mesh([
            pyrender.Primitive(
                positions=self._world_rays[name].T,
                indices=mesh_faces.T,
                color_0=np.ones((self._world_rays[name].shape[1], 4)),
                mode=pyrender.constants.GLTF.TRIANGLES,
            )
        ])
        self._bg_node[name] = pyrender.Node(
            name='bg_{}'.format(name),  # project camera view to 3D
            mesh=mesh,
            translation=camera_param.get_position()[:, 0],
            rotation=camera_param.get_quaternion()[:, 0])
        self._scene.add_node(self._bg_node[name])

    def _get_homogeneous_image_coords(self, camera_param, get_mesh=False):
        cam_w = camera_param.get_width()
        cam_h = camera_param.get_height()

        xx, yy = np.meshgrid(np.arange(cam_w), np.arange(cam_h))
        coords = np.stack(
            (xx.reshape(-1), yy.reshape(-1), np.ones_like(xx).reshape(-1)),
            axis=0)

        if not get_mesh:
            return coords
        else:
            upper = np.array([[0, 0], [0, 1], [1, 1]])
            lower = np.array([[0, 0], [1, 1], [1, 0]])
            mesh_tri = []

            # FIXME TODO: vectorize this double for-loop
            logging.debug('Homogeneous coordinate computation not vectorized')
            for i in range(0, cam_h - 1):
                for j in range(0, cam_w - 1):
                    c = np.array([i, j])
                    mesh_tri.append(
                        np.ravel_multi_index((c + upper).T, (cam_h, cam_w)))
                    mesh_tri.append(
                        np.ravel_multi_index((c + lower).T, (cam_h, cam_w)))
            mesh_tri = np.stack(mesh_tri, axis=1)

            return coords, mesh_tri

    @property
    def bg_mesh_names(self) -> List[str]:
        """ Names of all background meshes in the scene. """
        return self._bg_node.keys()

    @property
    def object_nodes(self) -> List[pyrender.Node]:
        """ Pyrender nodes of all virtual objects added to the scene. """
        return self._object_nodes

    @property
    def config(self) -> Dict:
        """ Configuration of the view synthesizer. """
        return self._config



import os
from typing import Optional, List, Any, Dict, Callable, Union, Tuple
import numpy as np
from scipy.interpolate import interp1d

from .core_utils import MultiSensor, LabelSearch, TopicNames
from ..utils import logging, misc


class Trace:
    """ This is a helper class for handling a trace in the dataset. It allows straightforward
    access to any pointer in the trace. A trace is separated into several segments based on
    timestamps (split to segments if the difference of two consecutive timestamps is larger than
    a value). Each segment contains frames that can be accessed through specifying frame index.
    Note that frame index used in VISTA is not necessarily the same as frame number in the dataset.
    Overall, we use ``segment_index``, ``frame_index``, ``frame_number``, ``timestamp`` in this
    class and there are several functions for conversion among them.
    The :class:`Trace` also performs filtering including,

        * Filter frames based on annotation label (`video_label.csv`) if there is any.
        * Filter by end-of-trace and time difference across consecutive frames.

    This class also handles sampling of pointer to the trace, i.e., sampling segment and frame
    index.

    Args:
        trace_path (str): Path to a trace.
        trace_config (dict): Configuration of the trace object.

    """
    DEFAULT_LABELS = [
        'day|night', 'dry|rain|snow',
        'local|residential|highway|unpaved|indoor', 'stable', '.*', '.*'
    ]
    RESET_CONFIG = {
        'default': {
            'n_bins': 100,
            'smoothing_factor': 0.001,
            'hist_power': 2.,
        },
        'segment_start': {
            'first_n_percent': 0.0,
        },
    }
    DEFAULT_CONFIG = {
        'reset_mode': 'default',
        'master_sensor': TopicNames.master_topic,
        'labels': DEFAULT_LABELS,
        'max_timestamp_diff_across_frames': 0.2,
        'road_width': 4,
    }

    def __init__(
        self, trace_path: str, trace_config: Optional[Dict] = dict()) -> None:
        self._trace_path: str = trace_path
        self._config: Dict = misc.merge_dict(trace_config, self.DEFAULT_CONFIG)

        # Get function representation of state information
        self._f_speed, self._f_curvature = self._get_states_func()

        # Divide trace to good segments based on video labels and timestamps
        self._multi_sensor: MultiSensor = MultiSensor(
            self._trace_path, self._config['master_sensor'])
        self._labels: LabelSearch = LabelSearch(*self._config['labels'])

        good_frames, good_timestamps = self._divide_to_good_segments()
        self._good_frames: Dict[str, List[int]] = good_frames
        self._good_timestamps: Dict[str, List[float]] = good_timestamps

        self._num_of_frames: int = np.sum([
            len(_v)
            for _v in self._good_frames[self._multi_sensor.master_sensor]
        ])

    def find_segment_reset(self) -> int:
        """ Sample a segment index based on number of frames in each segment. Segments with more
        frames will be sampled with a higher probability.

        Returns:
            int: Index of a segment.

        """
        segment_reset_probs = np.zeros(
            len(self._good_frames[self._multi_sensor.master_sensor]))
        for i in range(segment_reset_probs.shape[0]):
            segment = self._good_frames[self._multi_sensor.master_sensor][i]
            segment_reset_probs[i] = len(segment)
        segment_reset_probs /= np.sum(segment_reset_probs)
        new_segment_index = self._rng.choice(segment_reset_probs.shape[0],
                                             p=segment_reset_probs)

        return new_segment_index

    def find_frame_reset(self, segment_index: int) -> int:
        """ Sample a frame index in a segment.

        Args:
            segment_index (int): Index of the segment to be sampled from.

        Returns:
            int: Frame index.

        Raises:
            NotImplementedError: Unrecognized trace reset mode.

        """
        # Compute sample probability
        timestamps = self.good_timestamps[
            self._multi_sensor.master_sensor][segment_index]

        if self._config['reset_mode'] == 'default':
            # bias toward large road curvature
            n_bins = Trace.RESET_CONFIG['default']['n_bins']
            smoothing_factor = Trace.RESET_CONFIG['default'][
                'smoothing_factor']
            hist_power = Trace.RESET_CONFIG['default']['hist_power']

            curvatures = np.abs(self.f_curvature(timestamps))
            curvatures = np.clip(curvatures, 0, 1 / 3.)
            hist, bin_edges = np.histogram(curvatures, n_bins, density=False)
            bins = np.digitize(curvatures, bin_edges, right=True)
            hist_density = hist / float(np.sum(hist))
            probs = 1.0 / (
                (hist_density[bins - 1] + smoothing_factor)**hist_power)
            probs /= np.sum(probs)
        elif self._config['reset_mode'] == 'uniform':  # uniform sampling
            n_timestamps = len(timestamps)
            probs = np.ones((n_timestamps, )) / n_timestamps
        elif self._config[
                'reset_mode'] == 'segment_start':  # bias toward the start of a segment
            first_n_percent = Trace.RESET_CONFIG['segment_start'][
                'first_n_percent']

            n_timestamps = len(timestamps)
            probs = np.zeros((n_timestamps, ))
            to_idx = max(int(first_n_percent * n_timestamps), 1)
            probs[:to_idx] = 1.
            probs /= np.sum(probs)
        else:
            raise NotImplementedError(
                f'Unrecognized trace reset mode {self._reset_mode}')

        # Sample frame index
        frame_index = self._rng.choice(probs.shape[0], p=probs)

        return frame_index

    def get_master_timestamp(
            self,
            segment_index: int,
            frame_index: int,
            check_end: Optional[bool] = False
    ) -> Union[float, Tuple[bool, float]]:
        """ Get timestamp from the master sensor.

        Args:
            segment_index (int): Segment index.
            frame_index (int): Frame index.
            check_end (bool): Whether to check if the given segment and frame index
                              exceed the end of the trace; default to False.

        Returns:
            Return a tuple (``bool_a``, ``float_a``), where ``bool_a`` is whether the given segment
            and frame index exceed the end of the trace and ``float_a`` is timestamp; otherwise
            timestamp (float).

        """
        master_name = self.multi_sensor.master_sensor
        if check_end:
            exceed_end = frame_index >= len(
                self.good_timestamps[master_name][segment_index])
            frame_index = -1 if exceed_end else frame_index
            return exceed_end, self.good_timestamps[master_name][
                segment_index][frame_index]
        else:
            return self.good_timestamps[master_name][segment_index][
                frame_index]

    def get_master_frame_number(
            self,
            segment_index: int,
            frame_index: int,
            check_end: Optional[bool] = False) -> Union[int, Tuple[bool, int]]:
        """ Get frame number from the master sensor.

        Args:
            segment_index (int): Segment index.
            frame_index (int): Frame index.
            check_end (bool): Whether to check if the given segment and frame index
                              exceed the end of the trace; default to False.

        Returns:
            Return a tuple (``bool_a``, ``int_a``), where ``bool_a`` is whether the given segment
            and frame index exceed the end of the trace and ``int_a`` is frame number; otherwise
            frame number (int).

        """
        master_name = self.multi_sensor.master_sensor
        if check_end:
            exceed_end = frame_index >= len(
                self.good_timestamps[master_name][segment_index])
            frame_index = -1 if exceed_end else frame_index
            return exceed_end, self.good_frames[master_name][segment_index][
                frame_index]
        else:
            return self.good_frames[master_name][segment_index][frame_index]

    def _divide_to_good_segments(
        self,
        min_speed: float = 2.5,
    ) -> Dict[str, List[int]]:
        """ Divide a trace into good segments based on video labels and time difference between
        consecutive frames. Note that only master sensor is used for the time difference check
        since every sensors may have triggering frequencies.

        Returns:
            Return a tuple (``dict_a``, ``dict_b``), where
            ``dict_a``: good frames for all sensors. Key is sensor name and value
                        is a list with each element as frame indices of a good
                        segment, i.e., a good frame number =
                        dict[sensor_name][which_good_segment][i].
            ``dict_b``: timestamps of good frames.

        """
        # Filter by video labels
        _, good_labeled_timestamps = self._labels.find_good_labeled_frames(
            self._trace_path)
        if good_labeled_timestamps is None:
            logging.warning('No video_label.csv')
            good_labeled_timestamps = np.array(
                self._multi_sensor.get_master_timestamps())

        good_speed_inds = self.f_speed(good_labeled_timestamps) > min_speed
        good_labeled_timestamps = good_labeled_timestamps[good_speed_inds]

        # Filter by end-of-trace and time difference across consecutive frames
        good_frames = {_k: [] for _k in self._multi_sensor.sensor_names}
        good_timestamps = {_k: [] for _k in self._multi_sensor.sensor_names}
        segment_start = 0
        for i in range(good_labeled_timestamps.shape[0]):
            trace_end = i == good_labeled_timestamps.shape[0] - 1
            if not trace_end:
                time_diff = good_labeled_timestamps[
                    i + 1] - good_labeled_timestamps[i]
                time_diff_too_large = time_diff >= self._config[
                    'max_timestamp_diff_across_frames']
            else:
                time_diff_too_large = False

            if trace_end or time_diff_too_large:
                good_segment_frames = self._multi_sensor.get_frames_from_times(
                    good_labeled_timestamps[segment_start:i])

                for k, v in good_segment_frames.items():
                    good_frames[k].append(v)

                    segment_timestamps = []
                    for frame_num in v:
                        segment_timestamps.append(
                            self._multi_sensor.get_time_from_frame_num(
                                k, frame_num))
                    good_timestamps[k].append(segment_timestamps)

                segment_start += i

        return good_frames, good_timestamps

    def _get_states_func(self):
        """ Read speed and IMU data from the dataset and perform 1D interpolation to convert
        discrete data points to continuous functions. Curvature is computed as yaw rate (from
        IMU) divided by speed.

        Returns:
            Return a tuple (``func_a``, ``func_b``), where ``func_a`` is 1D interpolation
            function for speed and ``func_b`` is 1D interpolation for curvature.

        """
        # Read from dataset
        speed = np.genfromtxt(os.path.join(self._trace_path,
                                           TopicNames.speed + '.csv'),
                              delimiter=',')
        imu = np.genfromtxt(os.path.join(self._trace_path,
                                         TopicNames.imu + '.csv'),
                            delimiter=',')

        # Obtain function representation of speed
        f_speed = interp1d(speed[:, 0], speed[:, 1], fill_value='extrapolate')

        # Obtain function representation of curvature
        timestamps = imu[:, 0]
        yaw_rate = imu[:, 6]
        curvature = yaw_rate / np.maximum(f_speed(timestamps), 1e-10)
        good_curvature_inds = np.abs(curvature) < 1 / 3.
        f_curvature = interp1d(timestamps[good_curvature_inds],
                               curvature[good_curvature_inds],
                               fill_value='extrapolate')

        return f_speed, f_curvature

    def set_seed(self, seed) -> None:
        """ Set random seed.

        Args:
            seed (int): Random seed.

        """
        self._seed = seed
        self._rng = np.random.default_rng(self.seed)

    @property
    def seed(self) -> int:
        """ Random seed for sampling pointer in :meth:`find_segment_reset`
            and :meth:`find_frame_reset`. """
        return self._seed

    @property
    def trace_path(self) -> str:
        """ Path to the directory that contains all data of this trace. """
        return self._trace_path

    @property
    def multi_sensor(self) -> MultiSensor:
        """ MultiSensor object. Please check the definition of this class. """
        return self._multi_sensor

    @property
    def good_frames(self) -> Dict[str, List[int]]:
        """ A dictionary of good frame numbers, where keys are sensor names and
            values are lists of frame numbers (int). """
        return self._good_frames

    @property
    def good_timestamps(self) -> Dict[str, List[float]]:
        """ A dictionary of good timestamps, where keys are sensor names and
            values are a list of timestamps (float). """
        return self._good_timestamps

    @property
    def num_of_frames(self) -> int:
        """ Number of good frames. """
        return self._num_of_frames

    @property
    def f_curvature(self) -> Callable:
        """ A 1D interpolation function for curvature of the ego-car. """
        return self._f_curvature

    @property
    def f_speed(self) -> Callable:
        """ A 1D interpolation function for speed of the ego-car. """
        return self._f_speed

    @property
    def reset_mode(self) -> str:
        """ Trace reset mode. """
        return self._reset_mode

    @reset_mode.setter
    def reset_mode(self, reset_mode):
        assert isinstance(reset_mode, str)
        self._reset_mode = reset_mode

    @property
    def road_width(self) -> float:
        """ Road width. """
        return self._config['road_width']

    @property
    def param_file(self) -> float:
        """ Path to the sensor parameters file for this trace. """
        return os.path.join(self._trace_path, "params.xml")

    def __repr__(self) -> str:
        trace_name = os.path.basename(self.trace_path)
        return '<{}: {}>'.format(self.__class__.__name__, trace_name)



from typing import List, Optional
import os
import copy
import numpy as np
from PIL import Image
import trimesh
import pyrender
from vista.utils import transform


class MeshLib(object):
    """ Handle all meshes of actors (as opposed to scene/background itself) in the scene.
    It basically reads in all meshes with .obj extension, calibrates the meshes such that
    they are centered at the origin, and convert them to ``pyrender.Mesh`` for later usage.
    Note that this class is written specifically for a set of meshes (`carpack01`) and thus
    if there is a custom set of meshes, you would need to change how to read from the root
    directory and to calibrate the mesh.

    Args:
        root_dirs (List[str]): A list of root directories that contains meshes.

    Raises:
        AttributeError: If there is no mesh in the directory.

    In the source code, there is a ``main`` function that uses this class independently that
    performs rendering on meshes in pyrender.

    """

    FORMATS = ['.obj']

    def __init__(self, root_dirs: List[str]):
        root_dirs = [root_dirs
                     ] if not isinstance(root_dirs, list) else root_dirs
        fpaths = []
        for root_dir in root_dirs:
            root_dir = os.path.abspath(os.path.expanduser(root_dir))
            for sdir in os.listdir(root_dir):
                sdir = os.path.join(root_dir, sdir)
                if not os.path.isdir(sdir):
                    continue
                fpath = os.listdir(sdir)
                for fm in self.FORMATS:  # elements in the front has higher priority
                    fp = [
                        _fp for _fp in fpath if os.path.splitext(_fp)[-1] == fm
                    ]
                    if len(fp) == 1:
                        fpath = fp[0]
                        break
                if isinstance(fpath, list):
                    print(f'Cannot find mesh with supported format in {sdir}')
                    continue
                fpath = os.path.join(sdir, fpath)
                fpaths.append(fpath)

        tmeshes = dict()
        for i, fpath in enumerate(fpaths):
            try:
                tm = trimesh.load(fpath)
                tm = list(
                    tm.geometry.values())  # convert from scene to trimesh
                tm, mesh_dim = self._calibrate_tm(tm)
                source = os.path.basename(
                    os.path.dirname(os.path.dirname(fpath)))
                body_images = dict()
                for color in [
                        'Black', 'Blue', 'Green', 'Red', 'White', 'Yellow',
                        'Grey'
                ]:
                    img_path = os.path.splitext(fpath)[0] + color + '.png'
                    if os.path.exists(img_path):
                        img = Image.open(img_path)
                        body_images[color] = img
                extra = {'body_images': body_images}
                tmeshes[i] = {
                    'fpath': fpath,
                    'tmesh': tm,
                    'mesh_dim': mesh_dim,
                    'source': source,
                    'extra': extra,
                }
            except:
                print(f'Failed to load/process mesh {fpath}')
        if len(tmeshes) == 0:
            raise AttributeError("No mesh in the directory")

        self._fpaths = fpaths
        self._tmeshes = tmeshes
        self._agents_meshes = []
        self._agents_meshes_dim = []
        self._mesh_node = pyrender.Node()

    def reset(self, n_agents: int, random: Optional[bool] = True) -> None:
        """ Reset agents meshes by sampling ``n_agents`` meshes from the mesh library.

        Args:
            n_agents (int): Number of agents.
            random (bool): Whether to randomly sample ``n_agents`` meshes from
                the entire mesh library; default is set to ``True``.

        """
        idcs = np.random.choice(
            self.n_tmeshes, n_agents) if random else np.arange(self.n_tmeshes)

        self._agents_meshes = []
        self._agents_meshes_dim = []
        for idx in idcs:
            mesh = self._tmesh2mesh(self._tmeshes[idx])
            self._agents_meshes.append(mesh)
            self._agents_meshes_dim.append(self._tmeshes[idx]['mesh_dim'])

    def _tmesh2mesh(self, tm):
        # NOTE:cannot make a copy to keep the original tmesh intact otherwise will cause memory leak
        tm_list = tm['tmesh']

        # randomization in tmesh object
        # wheel, glass, optic, body = range(len(tm_list))
        body = np.argmax([v.triangles.shape[0] for v in tm_list
                          ])  # NOTE: hacky way to get body mesh
        color = np.random.choice(list(tm['extra']['body_images'].keys()))

        tm_list[body].visual.material.image = tm['extra']['body_images'][color]
        Ns = np.random.randint(10, 300)  # specular highlight (0-1000)
        tm_list[body].visual.material.kwargs.update(Ns=Ns)

        # convert mesh color from rgb to bgr
        for i in range(len(tm_list)):
            img = tm_list[i].visual.material.image
            if img is not None:
                tm_list[i].visual.material.image = Image.merge(
                    'RGB',
                    img.split()[::-1])

        # randomization in mesh object
        mesh = pyrender.Mesh.from_trimesh(tm_list)
        intensity = np.random.uniform(
            0., 1.)  # don't change base color; only change intensity

        mesh.primitives[body].material.baseColorFactor = np.array([intensity] *
                                                                  3 + [1.])
        mesh.primitives[body].material.metallicFactor = np.random.uniform(
            0.9, 1.0)
        mesh.primitives[body].material.roughnessFactor = np.random.uniform(
            0.0, 0.3)

        return mesh

    def _calibrate_tm(self, tm):
        # calibrate mesh; shift and scale assuming car dimension (width, length) is roughly (2, 4)
        all_pts = np.concatenate([np.array(_tm.vertices) for _tm in tm],
                                 axis=0)
        pts_min = all_pts.min(0)
        pts_max = all_pts.max(0)
        pts_range = pts_max - pts_min
        pts_midpoint = (pts_min + pts_max) / 2.
        xzy_shift = [-pts_midpoint[0], -pts_min[1],
                     -pts_midpoint[2]]  # on the ground
        rot = [0., np.pi, 0.]  # calibrate heading
        scale = 1. / (pts_range[0] / 2)  # assume all car width = 2
        for i in range(len(tm)):
            tr = transform.vec2mat(xzy_shift, rot)
            tm[i].apply_transform(tr)
            tm[i].apply_scale(scale)
        mesh_dim = [pts_range[0] * scale, pts_range[2] * scale]

        return tm, mesh_dim

    @property
    def fpaths(self) -> List[str]:
        """ Paths to all meshes. """
        return self._fpaths

    @property
    def tmeshes(self) -> List:
        """ A list of trimesh objects. """
        return self._tmeshes

    @property
    def n_tmeshes(self) -> int:
        """ Number of trimeshes. """
        return len(self._tmeshes)

    @property
    def agents_meshes(self) -> List[pyrender.Mesh]:
        """ A list of meshes for all agents. """
        return self._agents_meshes

    @property
    def agents_meshes_dim(self) -> List[List[float]]:
        """ The dimensions (width, length) of agents' meshes. """
        return self._agents_meshes_dim


def main():
    os.environ['PYOPENGL_PLATFORM'] = 'egl'
    import cv2
    import argparse
    import pdb
    from vista.entities.sensors.camera_utils import CameraParams

    parser = argparse.ArgumentParser(description='Run meshlib test.')
    parser.add_argument('--mesh-dir',
                        type=str,
                        nargs='+',
                        help='Directory to meshes.')
    parser.add_argument('--out-dir',
                        type=str,
                        help='Directory to rendered outputs.')
    parser.add_argument('--rig-path', type=str, help='Path to the rig file.')
    parser.add_argument('--repeat',
                        type=int,
                        default=1,
                        help='Number of repeat to render the same mesh.')
    parser.add_argument('--show',
                        action='store_true',
                        default=False,
                        help='Show image.')
    parser.add_argument('--dump-single',
                        action='store_true',
                        default=False,
                        help='Dump every single image.')
    parser.add_argument('--dump-merged',
                        action='store_true',
                        default=False,
                        help='Dump merged image.')
    args = parser.parse_args()

    if not os.path.isdir(args.out_dir) and (args.dump_single
                                            or args.dump_merged):
        os.makedirs(args.out_dir)

    mesh_lib = MeshLib(args.mesh_dir)

    scene = pyrender.Scene(ambient_light=[.5, .5, .5],
                           bg_color=[255, 255, 255])
    light = pyrender.DirectionalLight([255, 255, 255], 5)
    scene.add(light)

    camera = CameraParams(args.rig_path, 'camera_front')
    camera.resize(200, 320)
    render_camera = pyrender.IntrinsicsCamera(fx=camera._fx,
                                              fy=camera._fy,
                                              cx=camera._cx,
                                              cy=camera._cy,
                                              znear=0.01,
                                              zfar=100000)
    scene.add(render_camera)

    renderer = pyrender.OffscreenRenderer(camera.get_width(),
                                          camera.get_height())

    all_colors = []
    for r in range(args.repeat):
        mesh_lib.reset(mesh_lib.n_meshes, random=False)
        all_colors.append([])
        for i in range(mesh_lib.n_meshes):
            trans = [0.5, -0.6, 3.2]
            theta = np.deg2rad(50)
            rot = np.array([0, np.sin(theta / 2.), 0, np.cos(theta / 2.)])
            mesh_node = mesh_lib.get_mesh_node(i, trans, rot)
            scene.add_node(mesh_node)

            color, depth = renderer.render(scene)
            if args.show:
                cv2.imshow('rendered RGB', color)
                cv2.waitKey(0)
            all_colors[-1].append(color)

            if args.dump_single:
                fpath = os.path.basename(os.path.dirname(
                    mesh_lib.fpaths[i])) + f'_{r:02d}.png'
                fpath = os.path.join(args.out_dir, fpath)
                cv2.imwrite(fpath, color)

            scene.remove_node(mesh_node)

    if False:  # row image
        row_imgs = [np.concatenate(vl, axis=0) for vl in all_colors]
        merged_img = np.concatenate(row_imgs, axis=1)
    else:
        row_imgs = [np.concatenate(vl, axis=1) for vl in all_colors]
        merged_img = np.concatenate(row_imgs, axis=0)
    if args.dump_merged:
        fpath = os.path.join(args.out_dir, 'merged.png')
        cv2.imwrite(fpath, merged_img)


if __name__ == '__main__':
    main()



""" Logging utilities of the simulator. """
import os
from typing import Union
import inspect
import logging
from functools import partial

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL


class CustomFormatter(logging.Formatter):
    """ A custom logger class inherited from python logging module. It basically
    define a custom formatting as ``<time>::<level>::<module>::<msg>`` with colored
    text (``DEBUG``: cyan, ``INFO``: regular, ``WARNING``: yellow, ``ERROR``: bold red,
    ``CRITICAL``: underlined bold red). Otherwise, it can be used as if using regular
    python logging.

    Example Usage::

        >>> from vista.utils import logging
        >>> logging.setLevel(logging.ERROR)
        >>> logging.debug('This is a debug message.')
        >>> logging.warning('This is a warning message.')

    """
    PURPLE = '\033[95m'
    RED = "\033[0;31m"
    BOLD_RED = "\033[1;31m"
    BLUE = "\033[0;34m"
    CYAN = "\033[0;36m"
    GREEN = "\033[0;32m"
    YELLOW = '\033[93m'
    BOLD = "\033[;1m"
    UNDERLINE = '\033[4m'
    ENDC = '\033[0;0m'
    format = '%(asctime)s::%(levelname)s::%(message)s'

    FORMATS = {
        logging.DEBUG: CYAN + format + ENDC,
        logging.INFO: format + ENDC,
        logging.WARNING: YELLOW + format + ENDC,
        logging.ERROR: BOLD_RED + format + ENDC,
        logging.CRITICAL: UNDERLINE + BOLD_RED + format + ENDC
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


# Default logging config
logger = logging.getLogger('Vista')
logger.setLevel(logging.WARNING)
ch = logging.StreamHandler()
ch.setFormatter(CustomFormatter())
logger.addHandler(ch)


def setLevel(level: Union[str, int]) -> None:
    """ Set the logging level of this handler. Check CustomFormatter
    for example usage.

    Args:
        level (str or int): Logging level.

    """
    logger.setLevel(level)


def base(msg: str, func: str):
    filename = os.path.splitext(inspect.stack()[1].filename)[0].split('/')
    idx = [i for i, v in enumerate(filename) if v == 'vista']
    if len(idx) > 0:
        idx = idx[-1]
        filename = '.'.join(filename[idx:])
    else:
        filename = filename[0]
    func_name = inspect.stack()[1].function
    caller_name = '.'.join([filename, func_name])
    func = getattr(logger, func)
    func('[{}] {}'.format(caller_name, msg))


error = partial(base, func='error')
warning = partial(base, func='warning')
info = partial(base, func='info')
debug = partial(base, func='debug')
critical = partial(base, func='critical')
